# This file is part of Xpra.
# Copyright (C) 2013 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time

from OpenGL import GL

from xpra.util.env import envbool
from xpra.util.str_fn import strtobytes
from xpra.buffers.membuf import get_membuf  # @UnresolvedImport pylint: disable=import-outside-toplevel
from xpra.log import Logger

log = Logger("opengl", "paint")

SAVE_BUFFERS = os.environ.get("XPRA_OPENGL_SAVE_BUFFERS", "")
if SAVE_BUFFERS not in ("png", "jpeg", ""):
    log.warn("Warning: invalid value for XPRA_OPENGL_SAVE_BUFFERS: must be 'png' or 'jpeg'")
    SAVE_BUFFERS = ""
pillow_major = 0
if SAVE_BUFFERS:
    from PIL import Image, ImageOps, __version__ as pil_version
    try:
        pillow_major = int(pil_version.split(".")[0])
    except ValueError:
        pass

zerocopy_upload = False
if envbool("XPRA_OPENGL_ZEROCOPY_UPLOAD", True):
    try:
        import OpenGL_accelerate  # @UnresolvedImport

        assert OpenGL_accelerate
    except ImportError:
        pass
    else:
        from OpenGL import version

        zerocopy_upload = version.__version__ == OpenGL_accelerate.__version__


def pixels_for_upload(img_data) -> tuple[str, [memoryview | bytes]]:
    # prepare the pixel buffer for upload:
    if isinstance(img_data, memoryview):
        if zerocopy_upload:
            return "zerocopy:memoryview", img_data.toreadonly()
        # not safe, make a copy :(
        return "copy:memoryview.tobytes", img_data.tobytes()
    if isinstance(img_data, bytes):
        if zerocopy_upload:
            # we can zerocopy if we wrap it:
            return "zerocopy:bytes-as-memoryview", memoryview(img_data).toreadonly()
        return "copy:bytes", img_data
    if hasattr(img_data, "raw"):
        return "zerocopy:mmap", img_data.raw
    # everything else: copy to bytes:
    return f"copy:bytes({type(img_data)})", strtobytes(img_data)


def set_alignment(width: int, rowstride: int, pixel_format: str) -> None:
    bytes_per_pixel = len(pixel_format)  # ie: BGRX -> 4, Y -> 1, YY -> 2
    # Compute alignment and row length
    row_length = 0
    alignment = 1
    for a in (2, 4, 8):
        # Check if we are a-aligned - ! (var & 0x1) means 2-aligned or better, 0x3 - 4-aligned and so on
        if (rowstride & a - 1) == 0:
            alignment = a
    # If number of extra bytes is greater than the alignment value,
    # then we also have to set row_length
    # Otherwise it remains at 0 (= width implicitly)
    if (rowstride - width * bytes_per_pixel) >= alignment:
        row_length = width + (rowstride - width * bytes_per_pixel) // bytes_per_pixel
    GL.glPixelStorei(GL.GL_UNPACK_ROW_LENGTH, row_length)
    GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, alignment)
    # self.gl_marker("set_alignment%s GL_UNPACK_ROW_LENGTH=%i, GL_UNPACK_ALIGNMENT=%i",
    #               (width, rowstride, pixel_format), row_length, alignment)


def upload_rgba_texture(texture: int, width: int, height: int, pixels) -> None:
    upload, pixel_data = pixels_for_upload(pixels)
    rgb_format = "RGBA"
    target = GL.GL_TEXTURE_RECTANGLE
    GL.glBindTexture(target, texture)
    set_alignment(width, width * 4, rgb_format)
    GL.glTexParameteri(target, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
    GL.glTexParameteri(target, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
    GL.glTexParameteri(target, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
    GL.glTexParameteri(target, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)
    GL.glTexImage2D(target, 0, GL.GL_RGBA8, width, height, 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, pixel_data)
    log("upload_rgba_texture %ix%i uploaded %i bytes of %s pixel data using %s",
        width, height, len(pixels), rgb_format, upload)
    GL.glBindTexture(target, 0)


def save_fbo(wid: int, fbo, texture, width: int, height: int, alpha=False, prefix="W", suffix="") -> None:
    target = GL.GL_TEXTURE_RECTANGLE
    GL.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, fbo)
    GL.glBindTexture(target, texture)
    GL.glFramebufferTexture2D(GL.GL_READ_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0, target, texture, 0)
    GL.glReadBuffer(GL.GL_COLOR_ATTACHMENT0)
    GL.glViewport(0, 0, width, height)
    size = width * height * 4
    membuf = get_membuf(size)
    GL.glGetTexImage(target, 0, GL.GL_BGRA, GL.GL_UNSIGNED_BYTE, membuf.get_mem_ptr())
    if pillow_major < 10:
        pixels = memoryview(membuf)
    else:
        pixels = memoryview(membuf).tobytes()
    img = Image.frombuffer("RGBA", (width, height), pixels, "raw", "BGRA", width * 4)
    img = ImageOps.flip(img)
    kwargs = {}
    if alpha or SAVE_BUFFERS == "jpeg":
        img = img.convert("RGB")
    if SAVE_BUFFERS == "jpeg":
        kwargs = {
            "quality": 20,
            "optimize": False,
        }
    t = time.time()
    tstr = time.strftime("%H-%M-%S", time.localtime(t))
    millis = (t * 1000) % 1000
    filename = f"./{prefix}{wid}-FBO-{tstr}.{millis:03}{suffix}.{SAVE_BUFFERS}"
    log("save_fbo: saving %4ix%-4i pixels, %7i bytes to %s", width, height, size, filename)
    img.save(filename, SAVE_BUFFERS, **kwargs)
    GL.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, 0)
    GL.glBindTexture(target, 0)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys


def main() -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("SystemTray-Test", "SystemTray Test"):
        from xpra.gtk.examples import tray
        return tray.main()


if __name__ == "__main__":
    v = main()
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from collections.abc import Sequence

from xpra.util.env import envbool
from xpra.os_util import getuid
from xpra.scripts.config import str_to_bool
from xpra.auth.sys_auth_base import SysAuthenticator, log

PAM_AUTH_SERVICE = os.environ.get("XPRA_PAM_AUTH_SERVICE", "login")
PAM_CHECK_ACCOUNT = envbool("XPRA_PAM_CHECK_ACCOUNT", False)


def check(username: str, password: str, service: str = PAM_AUTH_SERVICE,
          check_account: bool = PAM_CHECK_ACCOUNT) -> bool:
    log("pam check(%r, ..)", username)
    from xpra.platform.pam import pam_session  # pylint: disable=import-outside-toplevel
    session = pam_session(username, password, service)
    if not session.start(password):
        return False
    try:
        success = session.authenticate()
        if success and check_account:
            success = session.check_account()
        return success
    except Exception as e:
        log.error("Error during pam authentication check")
        log.estr(e)
        return False
    finally:
        try:
            session.close()
        except Exception:
            log("error closing session %s", session, exc_info=True)


class Authenticator(SysAuthenticator):
    CLIENT_USERNAME = getuid() == 0

    def __init__(self, **kwargs):
        self.service = kwargs.pop("service", PAM_AUTH_SERVICE)
        self.check_account = str_to_bool(kwargs.pop("check-account", PAM_CHECK_ACCOUNT))
        super().__init__(**kwargs)

    def check_password(self, password: str) -> bool:
        log("pam.check_password(..) pw=%s", self.pw)
        if self.pw is None:
            return False
        return check(self.username, password, self.service, self.check_account)

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        self.req_xor(digests)
        return super().do_get_challenge(["xor"])

    def __repr__(self):
        return "PAM"


def main(args) -> int:
    if len(args) != 3:
        print("invalid number of arguments")
        print("usage:")
        print(f"{args[0]} username password")
        return 1
    username = args[1]
    a = Authenticator(username=username)
    if a.check(args[2]):
        print("success")
        return 0
    print("failed")
    return -1


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import hashlib
import os.path
from typing import Any, Optional

from xpra.audio.pulseaudio.common_util import get_pulse_server_x11_property, get_pulse_id_x11_property
from xpra.os_util import WIN32, OSX
from xpra.util.io import which
from xpra.util.str_fn import strtobytes, bytestostr

from xpra.log import Logger

log = Logger("audio")

pactl_bin = None
has_pulseaudio = None


def get_pactl_bin() -> str:
    global pactl_bin
    if pactl_bin is None:
        if WIN32 or OSX:
            pactl_bin = ""
        else:
            pactl_bin = which("pactl")
    return pactl_bin


def pactl_output(log_errors=True, *pactl_args) -> tuple[int, Any, Any]:
    pactl_bin = get_pactl_bin()
    if not pactl_bin:
        return -1, None, None
    # ie: "pactl list"
    cmd = [pactl_bin] + list(pactl_args)
    # force "C" locale so that we can parse the output as expected
    env = os.environ.copy()
    env["LC_ALL"] = "C"
    env.pop("DISPLAY", None)
    try:
        import subprocess
        log(f"running {cmd!r} with env={env}")
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
        from xpra.util.child_reaper import getChildReaper
        procinfo = getChildReaper().add_process(process, "pactl", cmd, True, True)
        log(f"waiting for {cmd!r} output")
        out, err = process.communicate()
        getChildReaper().add_dead_process(procinfo)
        code = process.wait()
        log(f"pactl_output{pactl_args} returned {code}")
        return code, out, err
    except Exception as e:
        if log_errors:
            log.error("failed to execute %s: %s", cmd, e)
        else:
            log("failed to execute %s: %s", cmd, e)
        return -1, None, None


def is_pa_installed() -> bool:
    pactl_bin = get_pactl_bin()
    log("is_pa_installed() pactl_bin=%s", pactl_bin)
    return bool(pactl_bin)


def has_pa() -> bool:
    global has_pulseaudio
    if has_pulseaudio is None:
        has_pulseaudio = get_pulse_server_x11_property() or is_pa_installed()
    return bool(has_pulseaudio)


def set_source_mute(device, mute=False) -> bool:
    code, out, err = pactl_output(True, "set-source-mute", device, str(int(mute)))
    log("set_source_mute: output=%s, err=%s", out, err)
    return code == 0


def set_sink_mute(device, mute=False) -> bool:
    code, out, err = pactl_output(True, "set-sink-mute", device, str(int(mute)))
    log("set_sink_mute: output=%s, err=%s", out, err)
    return code == 0


def get_pactl_info_line(prefix) -> str:
    if not has_pa():
        return ""
    code, out, err = pactl_output(False, "info")
    if code != 0:
        log.warn("Warning: failed to query pulseaudio using 'pactl info'")
        if err:
            for x in err.splitlines():
                log.warn(" %s", bytestostr(x))
        return ""
    stat = ""
    for line in bytestostr(out).splitlines():
        if line.startswith(prefix):
            stat = line[len(prefix):].strip()
            break
    log("get_pactl_info_line(%s)=%s", prefix, stat)
    return stat


def get_default_sink() -> str:
    return get_pactl_info_line("Default Sink:")


def get_pactl_server() -> str:
    return get_pactl_info_line("Server String:")


def get_pulse_cookie_hash() -> bytes:
    v = get_pactl_info_line("Cookie:")
    return strtobytes(hashlib.sha256(strtobytes(v)).hexdigest())


def get_pulse_server(may_start_it=True) -> str:
    xp = get_pulse_server_x11_property()
    if xp or not may_start_it:
        return bytestostr(xp or b"")
    return get_pactl_server()


def get_pulse_id() -> str:
    return get_pulse_id_x11_property()


def get_pa_device_options(monitors=False, input_or_output=None, ignored_devices=("bell-window-system",)):
    """
    Finds the list of devices, monitors=False allows us to filter out monitors
    (which could create audio loops if we use them)
    set input_or_output=True to get inputs only
    set input_or_output=False to get outputs only
    set input_or_output=None to get both
    Same goes for monitors (False|True|None)
    Returns a dict() with the PulseAudio name as key and a description as value
    """
    if WIN32 or OSX:
        return {}
    status, out, _ = pactl_output(False, "list")
    if status != 0 or not out:
        return {}
    return do_get_pa_device_options(out, monitors, input_or_output, ignored_devices)


def do_get_pa_device_options(pactl_list_output, monitors=False, input_or_output=None,
                             ignored_devices=("bell-window-system",)):
    def are_properties_acceptable(name: Optional[str], device_class: Optional[str],
                                  monitor_of_sink: Optional[str]) -> bool:
        if (name is None) or (device_class is None and monitor_of_sink is None):
            return False
        if name in ignored_devices:
            return False
        # Verify against monitor flag if set:
        if monitors is not None:
            if device_class is not None:
                is_monitor = device_class == '"monitor"'
            else:  # monitor_of_sink is not None
                is_monitor = monitor_of_sink != "n/a"
            if is_monitor != monitors:
                return False
        # Verify against input flag (if set):
        if input_or_output is not None:
            is_input = name.lower().find("input") >= 0
            if is_input is True and input_or_output is False:
                return False
            is_output = name.lower().find("output") >= 0
            if is_output is True and input_or_output is True:
                return False
        return True

    name: Optional[str] = None
    device_class: Optional[str] = None
    monitor_of_sink: Optional[str] = None
    device_description: Optional[str] = None
    devices: dict[str, str] = {}
    for line in bytestostr(pactl_list_output).splitlines():
        if not line.startswith(" ") and not line.startswith("\t"):  # clear vars when we encounter a new section
            if are_properties_acceptable(name, device_class, monitor_of_sink):
                assert isinstance(name, str)
                if not device_description:
                    device_description = name
                devices[name] = device_description
            name = None
            device_class = None
            monitor_of_sink = None
            device_description = None
        line = line.strip()
        if line.startswith("Name: "):
            name = line[len("Name: "):]
        if line.startswith("device.class = "):
            device_class = line[len("device-class = "):]
        if line.startswith("Monitor of Sink: "):
            monitor_of_sink = line[len("Monitor of Sink: "):]
        if line.startswith("device.description = "):
            device_description = line[len("device.description = "):].strip('"')
    return devices


def get_info() -> dict[str, Any]:
    i = 0
    dinfo = {}
    status, out, _ = pactl_output(False, "list")
    if status == 0 and out:
        for monitors in (True, False):
            for io in (True, False):
                devices = do_get_pa_device_options(out, monitors, io)
                for d, name in devices.items():
                    dinfo[bytestostr(d)] = bytestostr(name)
                    i += 1
    info = {
        "device": dinfo,
        "devices": i,
        "pulseaudio": {
            "wrapper": "pactl",
            "found": bool(has_pa()),
            "id": get_pulse_id(),
            "server": get_pulse_server(False),
            "cookie-hash": get_pulse_cookie_hash(),
        }
    }
    log("pulseaudio_pactl_util.get_info()=%s", info)
    return info


def main():
    from xpra.util.str_fn import print_nested_dict
    from xpra.util.io import load_binary_file
    if "-v" in sys.argv:
        log.enable_debug()
        sys.argv.remove("-v")
    if len(sys.argv) > 1:
        for filename in sys.argv[1:]:
            if not os.path.exists(filename):
                log.warn("file argument '%s' does not exist, ignoring", filename)
                continue
            data = load_binary_file(filename)
            devices = do_get_pa_device_options(data, True, False)
            log.info("%s devices found in '%s'", len(devices), filename)
            print_nested_dict(devices)
        return

    i = get_info()
    print_nested_dict(i)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

import os
from time import sleep
from typing import Any
from collections.abc import Callable

from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.scripts.config import parse_with_unit
from xpra.util.stats import std_unit
from xpra.net.common import PacketType
from xpra.os_util import POSIX, gi_import
from xpra.util.io import livefds
from xpra.util.pysystem import detect_leaks
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool
from xpra.log import Logger

log = Logger("network")
bandwidthlog = Logger("network", "bandwidth")
eventslog = Logger("events")

GLib = gi_import("GLib")

DETECT_MEMLEAKS = envint("XPRA_DETECT_MEMLEAKS", 0)
DETECT_FDLEAKS = envbool("XPRA_DETECT_FDLEAKS", False)

MIN_BANDWIDTH_LIMIT = envint("XPRA_MIN_BANDWIDTH_LIMIT", 1024 * 1024)
MAX_BANDWIDTH_LIMIT = envint("XPRA_MAX_BANDWIDTH_LIMIT", 10 * 1024 * 1024 * 1024)
CPUINFO = envbool("XPRA_CPUINFO", False)


class NetworkStateServer(StubServerMixin):
    """
    Mixin for adding client / network state monitoring functions:
    - ping and echo
    - bandwidth management
    - leak detection (file descriptors, memory)
    """

    def __init__(self):
        self.pings = False
        self.ping_timer: int = 0
        self.bandwidth_limit = 0
        self.bandwidth_detection = False
        self.mem_bytes = 0
        self.cpu_info: dict = {}
        self.print_memleaks: Callable | None = None
        self.fds: set[int] = set()

    def init(self, opts) -> None:
        self.pings = opts.pings
        self.bandwidth_limit = parse_with_unit("bandwidth-limit", opts.bandwidth_limit) or 0
        self.bandwidth_detection = opts.bandwidth_detection
        bandwidthlog("bandwidth-limit(%s)=%s", opts.bandwidth_limit, self.bandwidth_limit)
        self.init_cpuinfo()

    def setup(self) -> None:
        self.init_leak_detection()
        if self.pings > 0:
            self.ping_timer = GLib.timeout_add(1000 * self.pings, self.send_ping)

    def threaded_setup(self) -> None:
        self.init_memcheck()

    def cleanup(self) -> None:
        pt = self.ping_timer
        if pt:
            self.ping_timer = 0
            GLib.source_remove(pt)
        pm = self.print_memleaks
        if pm:
            pm()

    def get_info(self, _source=None) -> dict[str, Any]:
        info = {
            "pings": self.pings,
            "bandwidth-limit": self.bandwidth_limit or 0,
        }
        if POSIX:
            info["load"] = tuple(int(x * 1000) for x in os.getloadavg())
        if self.mem_bytes:
            info["total-memory"] = self.mem_bytes
        if self.cpu_info:
            info["cpuinfo"] = {k: v for k, v in self.cpu_info.items() if k != "python_version"}
        return info

    def get_server_features(self, _source) -> dict[str, Any]:
        return {
            "network": {
                "bandwidth-limit": self.bandwidth_limit or 0,
            }
        }

    def init_leak_detection(self) -> None:
        if DETECT_MEMLEAKS:
            pm = detect_leaks()
            self.print_memleaks = pm
            if bool(self.print_memleaks):
                def leak_thread():
                    while not self._closing:
                        pm()
                        sleep(DETECT_MEMLEAKS)

                from xpra.util.thread import start_thread  # pylint: disable=import-outside-toplevel
                start_thread(leak_thread, "leak thread", daemon=True)
        if DETECT_FDLEAKS:
            self.fds = livefds()

            def print_fds():
                fds = livefds()
                newfds = fds - self.fds
                self.fds = fds
                log.info("print_fds() new fds=%s (total=%s)", newfds, len(fds))
                return True

            GLib.timeout_add(10, print_fds)

    def init_memcheck(self) -> None:
        # verify we have enough memory:
        if POSIX and self.mem_bytes == 0:
            try:
                self.mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')  # e.g. 4015976448
                LOW_MEM_LIMIT = 1024 * 1024 * 1024
                if self.mem_bytes <= LOW_MEM_LIMIT:
                    log.warn("Warning: only %iMB total system memory available", self.mem_bytes // (1024 ** 2))
                    log.warn(" this may not be enough to run a server")
                else:
                    log.info("%.1fGB of system memory", self.mem_bytes / (1024.0 ** 3))
            except Exception:
                pass

    def init_cpuinfo(self) -> None:
        if not CPUINFO:
            return
        # this crashes if not run from the UI thread!
        try:
            from cpuinfo import get_cpu_info
        except ImportError as e:
            log("no cpuinfo: %s", e)
            return
        self.cpu_info = get_cpu_info()
        if self.cpu_info:
            c = typedict(self.cpu_info)
            count = c.intget("count", 0)
            brand = c.strget("brand")
            if count > 0 and brand:
                log.info("%ix %s", count, brand)

    def _process_connection_data(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if ss:
            ss.update_connection_data(packet[1])

    def _process_bandwidth_limit(self, proto, packet: PacketType) -> None:
        log("_process_bandwidth_limit(%s, %s)", proto, packet)
        ss = self.get_server_source(proto)
        if not ss:
            return
        bandwidth_limit = packet[1]
        if not isinstance(bandwidth_limit, int):
            raise TypeError("bandwidth-limit must be an integer, not %s" % type(bandwidth_limit))
        if (self.bandwidth_limit and bandwidth_limit >= self.bandwidth_limit) or bandwidth_limit <= 0:
            bandwidth_limit = self.bandwidth_limit or 0
        if ss.bandwidth_limit == bandwidth_limit:
            # unchanged
            log("bandwidth limit unchanged: %s", std_unit(bandwidth_limit))
            return
        if bandwidth_limit < MIN_BANDWIDTH_LIMIT:
            log.warn("Warning: bandwidth limit requested is too low (%s)", std_unit(bandwidth_limit))
            bandwidth_limit = MIN_BANDWIDTH_LIMIT
        if bandwidth_limit >= MAX_BANDWIDTH_LIMIT:
            log("bandwidth limit over maximum, using no-limit instead")
            bandwidth_limit = 0
        ss.bandwidth_limit = bandwidth_limit
        # we can't assume to have a full ClientConnection object:
        client_id = getattr(ss, "counter", "")
        if bandwidth_limit == 0:
            bandwidthlog.info("bandwidth-limit restrictions removed for client %s", client_id)
        else:
            bandwidthlog.info("bandwidth-limit changed to %sbps for client %s", std_unit(bandwidth_limit), client_id)

    def send_ping(self) -> bool:
        from xpra.server.source.networkstate import NetworkStateMixin
        for ss in self._server_sources.values():
            if ss.suspended or ss.is_closed() or not ss.startup_completed:
                continue
            if isinstance(ss, NetworkStateMixin):
                ss.ping()
        return True

    def _process_ping_echo(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if ss:
            ss.process_ping_echo(packet)

    def _process_ping(self, proto, packet: PacketType) -> None:
        time_to_echo = packet[1]
        sid = ""
        if len(packet) >= 4:
            sid = packet[3]
        ss = self.get_server_source(proto)
        if ss:
            ss.process_ping(time_to_echo, sid)

    def _process_suspend(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        eventslog("suspend(%s) source=%s", packet[1:], ss)
        if ss:
            ss.suspend()

    def _process_resume(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        eventslog("resume(%s) source=%s", packet[1:], ss)
        if ss:
            ss.resume()

    def init_packet_handlers(self) -> None:
        self.add_packets(
            "ping", "ping_echo",
            "connection-data", "bandwidth-limit",
        )
        self.add_packets(
            "suspend", "resume",
            main_thread=True,
        )



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
import os.path
from threading import Lock
from typing import Any
from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.util.env import envint, envbool
from xpra.util.thread import start_thread
from xpra.server.background_worker import add_work_item
from xpra.log import Logger

GLib = gi_import("GLib")
Gio = gi_import("Gio")

log = Logger("menu")

MENU_WATCHER = envbool("XPRA_MENU_WATCHER", True)
MENU_RELOAD_DELAY = envint("XPRA_MENU_RELOAD_DELAY", 5)
EXPORT_MENU_DATA = envbool("XPRA_EXPORT_MENU_DATA", True)


def noicondata(menu_data: dict) -> dict:
    newdata = {}
    for k, v in menu_data.items():
        if k in ("IconData", b"IconData"):
            continue
        if isinstance(v, dict):
            newdata[k] = noicondata(v)
        else:
            newdata[k] = v
    return newdata


class MenuProvider:
    __slots__ = (
        "dir_watchers", "menu_reload_timer",
        "on_reload", "menu_data", "desktop_sessions", "load_lock",
    )

    def __init__(self):
        self.dir_watchers: dict[str, Any] = {}
        self.menu_reload_timer = 0
        self.on_reload: list[Callable] = []
        self.menu_data: dict[str, Any] | None = None
        self.desktop_sessions: dict[str, Any] | None = None
        self.load_lock = Lock()

    def setup(self) -> None:
        if not EXPORT_MENU_DATA:
            return
        if MENU_WATCHER:
            self.setup_menu_watcher()
        self.load_menu_data()

    def cleanup(self) -> None:
        self.on_reload = []
        self.cancel_menu_reload()
        self.cancel_dir_watchers()

    def setup_menu_watcher(self) -> None:
        try:
            self.do_setup_menu_watcher()
        except Exception as e:
            log("threaded_setup()", exc_info=True)
            log.error("Error setting up menu watcher:")
            log.estr(e)

    def do_setup_menu_watcher(self) -> None:
        def directory_changed(*args) -> None:
            log(f"directory_changed{args}")
            self.schedule_menu_reload()

        from xpra.platform.paths import get_system_menu_dirs
        try:
            from watchdog.events import (
                FileSystemEvent, FileSystemEventHandler,
                EVENT_TYPE_MOVED, EVENT_TYPE_DELETED, EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED,
                # unused: EVENT_TYPE_CLOSED, EVENT_TYPE_OPENED
            )
            DIR_CHANGE_EVENT_TYPES = (EVENT_TYPE_MOVED, EVENT_TYPE_DELETED, EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED)
            from watchdog.observers import Observer

            class MenuDirEventHandler(FileSystemEventHandler):
                def on_any_event(self, event: FileSystemEvent) -> None:
                    if event.event_type in DIR_CHANGE_EVENT_TYPES:
                        directory_changed(event)

            event_handler = MenuDirEventHandler()
            observer = Observer()
            # alias so we can use the same method call as Gio to stop it:
            observer.cancel = observer.stop
            for menu_dir in get_system_menu_dirs():
                dir_path = os.path.abspath(menu_dir)
                if not os.path.exists(dir_path) or not os.path.isdir(dir_path):
                    continue
                if dir_path not in self.dir_watchers:
                    observer.schedule(event_handler, dir_path, recursive=True)
                    self.dir_watchers[dir_path] = observer
            log(f"using watchdog library: {observer} and {event_handler}")
            observer.start()
            watcher = "watchdog library"
        except ImportError as e:
            log(f"watchdog library not found: {e}, using Gio instead")
            for menu_dir in get_system_menu_dirs():
                if menu_dir not in self.dir_watchers:
                    gfile = Gio.File.new_for_path(menu_dir)
                    monitor = gfile.monitor_directory(Gio.FileMonitorFlags.NONE, None)
                    monitor.connect("changed", directory_changed)
                    self.dir_watchers[menu_dir] = monitor
            watcher = "Gio file monitor"
        if self.dir_watchers:
            log.info("watching for applications menu changes in:")
            for wd in self.dir_watchers.keys():
                log.info(f" {wd!r}")
            log.info(f"using {watcher!r}")

    def cancel_dir_watchers(self) -> None:
        dw = self.dir_watchers
        self.dir_watchers = {}
        for monitor in set(dw.values()):
            monitor.cancel()

    def load_menu_data(self, force_reload: bool = False) -> None:
        # start loading in a thread,
        # as this may take a while and
        # so server startup can complete:
        def load() -> None:
            try:
                self.get_menu_data(force_reload)
                self.get_desktop_sessions()
            except ImportError as e:
                log.warn("Warning: cannot load menu data")
                log.warn(f" {e}")
            except Exception:
                log.error("Error loading menu data", exc_info=True)
            finally:
                self.clear_cache()

        start_thread(load, "load-menu-data", True)

    def get_menu_data(self, force_reload=False, remove_icons=False, wait=True) -> dict[str, Any]:
        log("get_menu_data%s", (force_reload, remove_icons, wait))
        if not EXPORT_MENU_DATA:
            return {}
        menu_data = self.menu_data
        if self.load_lock.acquire(wait):  # pylint: disable=consider-using-with
            menu_data = self.menu_data
            try:
                if menu_data is None or force_reload:
                    from xpra.platform.menu_helper import load_menu  # pylint: disable=import-outside-toplevel
                    menu_data = self.menu_data = load_menu()
                    add_work_item(self.got_menu_data)
            finally:
                self.load_lock.release()
        if remove_icons and self.menu_data:
            menu_data = noicondata(self.menu_data)
        return menu_data or {}

    def got_menu_data(self) -> bool:
        log("got_menu_data(..) on_reload=%s", self.on_reload)
        for cb in self.on_reload:
            cb(self.menu_data)
        return False

    def clear_cache(self) -> None:
        from xpra.platform.menu_helper import clear_cache  # pylint: disable=import-outside-toplevel
        log("%s()", clear_cache)
        clear_cache()

    def cancel_menu_reload(self) -> None:
        xmrt = self.menu_reload_timer
        if xmrt:
            self.menu_reload_timer = 0
            GLib.source_remove(xmrt)

    def schedule_menu_reload(self) -> None:
        self.cancel_menu_reload()
        self.menu_reload_timer = GLib.timeout_add(MENU_RELOAD_DELAY * 1000, self.menu_reload)

    def menu_reload(self) -> bool:
        self.menu_reload_timer = 0
        log("menu_reload()")
        self.load_menu_data(True)
        return False

    def get_menu_icon(self, category_name: str, app_name: str) -> tuple[str, bytes]:
        menu_data = self.get_menu_data()
        if not menu_data:
            return "", b""
        category = menu_data.get(category_name)
        if not category:
            log("get_menu_icon: invalid menu category '%s'", category_name)
            return "", b""
        if not app_name:
            return category.get("IconType"), category.get("IconData")
        entries = category.get("Entries")
        if not entries:
            log("get_menu_icon: no entries for category '%s'", category_name)
            return "", b""
        app = entries.get(app_name)
        if not app:
            log("get_menu_icon: no matching application for '%s' in category '%s'",
                app_name, category_name)
            return "", b""
        return app.get("IconType", ""), app.get("IconData", b"")

    def get_desktop_sessions(self, force_reload: bool = False, remove_icons: bool = False) -> dict[str, Any]:
        if force_reload or self.desktop_sessions is None:
            from xpra.platform.menu_helper import load_desktop_sessions  # pylint: disable=import-outside-toplevel
            self.desktop_sessions = load_desktop_sessions()
        desktop_sessions = self.desktop_sessions
        if remove_icons:
            desktop_sessions = noicondata(desktop_sessions)
        return desktop_sessions

    def get_desktop_menu_icon(self, sessionname: str) -> tuple[str, bytes]:
        desktop_sessions = self.get_desktop_sessions(False) or {}
        de = desktop_sessions.get(sessionname, {})
        return de.get("IconType", ""), de.get("IconData")

    def get_info(self, _proto) -> dict[str, Any]:
        return self.get_menu_data(remove_icons=True)


singleton: MenuProvider | None = None


def get_menu_provider() -> MenuProvider:
    global singleton
    if singleton is None:
        singleton = MenuProvider()
    return singleton



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from typing import Any

from xpra.os_util import gi_import
from xpra.util.system import get_generic_os_name
from xpra.util.io import load_binary_file
from xpra.platform.paths import get_icon, get_icon_filename
from xpra.gtk.gobject import no_arg_signal, one_arg_signal
from xpra.gtk.error import xlog
from xpra.x11.common import get_wm_name
from xpra.x11.models.model_stub import WindowModelStub
from xpra.x11.bindings.window import X11WindowBindings
from xpra.x11.gtk.damage import WindowDamageHandler
from xpra.x11.gtk.bindings import add_event_receiver, remove_event_receiver
from xpra.x11.bindings.randr import RandRBindings
from xpra.log import Logger

X11Window = X11WindowBindings()
RandR = RandRBindings()

eventlog = Logger("server", "window", "events")
geomlog = Logger("server", "window", "geometry")
iconlog = Logger("icon")

GObject = gi_import("GObject")
GLib = gi_import("GLib")


class DesktopModelBase(WindowModelStub, WindowDamageHandler):
    __common_gsignals__ = {}
    __common_gsignals__ |= WindowDamageHandler.__common_gsignals__
    __common_gsignals__ |= {
        "resized": no_arg_signal,
        "client-contents-changed": one_arg_signal,
        "motion": one_arg_signal,
        "x11-motion-event": one_arg_signal,
        "x11-property-notify-event": one_arg_signal,
    }

    __gproperties__ = {
        "iconic": (
            GObject.TYPE_BOOLEAN,
            "ICCCM 'iconic' state -- any sort of 'not on desktop'.", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "focused": (
            GObject.TYPE_BOOLEAN,
            "Is the window focused", "",
            False,
            GObject.ParamFlags.READWRITE,
        ),
        "size-hints": (
            GObject.TYPE_PYOBJECT,
            "Client hints on constraining its size", "",
            GObject.ParamFlags.READABLE,
        ),
        "wm-name": (
            GObject.TYPE_PYOBJECT,
            "The name of the window manager or session manager", "",
            GObject.ParamFlags.READABLE,
        ),
        "title": (
            GObject.TYPE_PYOBJECT,
            "The name of this desktop or monitor", "",
            GObject.ParamFlags.READABLE,
        ),
        "icons": (
            GObject.TYPE_PYOBJECT,
            "The icon of the window manager or session manager", "",
            GObject.ParamFlags.READABLE,
        ),
    }

    _property_names = [
        "client-machine", "window-type",
        "shadow", "size-hints", "class-instance",
        "focused", "title", "depth", "icons",
        "content-type",
        "set-initial-position",
    ]
    _dynamic_property_names = ["size-hints", "title", "icons"]

    def __init__(self):
        root_xid = X11Window.get_root_xid()
        WindowDamageHandler.__init__(self, root_xid)
        WindowModelStub.__init__(self)
        self.update_wm_name()
        self.update_icon()
        self._depth = 24
        self.resize_timer = 0
        self.resize_value = (1, 1)

    def setup(self) -> None:
        WindowDamageHandler.setup(self)
        self._depth = X11Window.get_depth(self.xid)
        X11Window.addDefaultEvents(self.xid)
        self._managed = True
        self._setup_done = True
        # listen for property changes on the root window:
        # (monitor mode can have up to 16 monitors, so we may have many event listeners
        #  so raise the limit to 20)
        add_event_receiver(self.xid, self, 20)

    def do_x11_property_notify_event(self, event) -> None:
        eventlog(f"do_x11_property_notify_event: {event.atom}")
        # update the wm-name (and therefore the window's "title")
        # whenever this property changes:
        if str(event.atom) == "_NET_SUPPORTING_WM_CHECK":
            if self.update_wm_name():
                self.update_icon()
                self.notify("title")

    def unmanage(self, exiting=False) -> None:
        remove_event_receiver(self.xid, self)
        WindowDamageHandler.destroy(self)
        WindowModelStub.unmanage(self, exiting)
        self.cancel_resize_timer()
        self._managed = False

    def update_wm_name(self) -> bool:
        wm_name = ""
        with xlog:
            wm_name = get_wm_name()  # pylint: disable=assignment-from-none
        iconlog("update_wm_name() wm-name=%s", wm_name)
        return self._updateprop("wm-name", wm_name)

    def update_icon(self) -> bool:
        wm_name = self.get_property("wm-name")
        if not wm_name:
            return False
        icon_name = get_icon_filename(wm_name.lower() + ".png")
        if not icon_name:
            return False
        try:
            from xpra.codecs.pillow.decoder import open_only  # pylint: disable=import-outside-toplevel
        except ImportError:
            iconlog("unable to get icon without pillow")
            return False
        icon_data = load_binary_file(icon_name)
        if not icon_data:
            raise ValueError(f"failed to load icon {icon_name!r}")
        try:
            img = open_only(icon_data, types=("png",))
            w, h = img.size
            img.close()
        except OSError:
            iconlog("failed to open window icon", exc_info=True)
            return False
        iconlog("Image(%s)=%s", icon_name, img)
        if not img:
            return False
        icon = (w, h, "png", icon_data)
        icons = (icon,)
        return self._updateprop("icons", icons)

    def uses_xshm(self) -> bool:
        return bool(self._xshm_handle)

    def get_default_window_icon(self, _size: int = 48) -> tuple[int, int, str, bytes] | None:
        icon_name = get_generic_os_name() + ".png"
        icon = get_icon(icon_name)
        if not icon:
            return None
        return icon.get_width(), icon.get_height(), "RGBA", icon.get_pixels()

    def get_title(self) -> str:
        return self.get_property("wm-name") or "xpra desktop"

    def get_property(self, prop: str) -> Any:
        if prop == "depth":
            return self._depth
        if prop == "title":
            return self.get_title()
        if prop == "client-machine":
            return socket.gethostname()
        if prop == "window-type":
            return ["NORMAL"]
        if prop == "shadow":
            return True
        if prop == "class-instance":
            return "xpra-desktop", "Xpra-Desktop"
        if prop == "content-type":
            return "desktop"
        if prop == "set-initial-position":
            return False
        return GObject.GObject.get_property(self, prop)

    def do_x11_damage_event(self, event) -> None:
        self.emit("client-contents-changed", event)

    def do_x11_motion_event(self, event) -> None:
        self.emit("motion", event)

    def resize(self, w: int, h: int) -> None:
        geomlog("resize(%i, %i)", w, h)
        if not RandR.has_randr():
            geomlog.error("Error: cannot honour resize request,")
            geomlog.error(" no RandR support on this display")
            return
        # FIXME: small race if the user resizes with randr,
        # at the same time as he resizes the window..
        self.resize_value = (w, h)
        if not self.resize_timer:
            self.resize_timer = GLib.timeout_add(250, self.do_resize)

    def do_resize(self) -> None:
        raise NotImplementedError

    def cancel_resize_timer(self) -> None:
        rt = self.resize_timer
        if rt:
            self.resize_timer = 0
            GLib.source_remove(rt)



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.util.objects import AtomicInteger
from xpra.util.str_fn import csv, bytestostr
from xpra.util.env import IgnoreWarningsContext
from xpra.common import gravity_str, WORKSPACE_UNSET
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.log import Logger

log = Logger("info")

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")


def slabel(text: str = "", tooltip: str = "", font: str = "") -> Gtk.Label:
    lbl = label(text, tooltip, font)
    lbl.set_selectable(True)
    return lbl


def dict_str(d) -> str:
    return "\n".join("%s : %s" % (k, v) for k, v in d.items())


def geom_str(geom) -> str:
    return "%ix%i at %i,%i" % (geom[2], geom[3], geom[0], geom[1])


def hsc(sc) -> str:
    # make the dict more human-readable
    ssc = dict((bytestostr(k), v) for k, v in sc.items())
    ssc.pop("gravity", None)
    return dict_str(ssc)


def get_window_state(w) -> str:
    state = []
    for s in (
            "fullscreen", "maximized",
            "above", "below", "shaded", "sticky",
            "skip-pager", "skip-taskbar",
            "iconified",
    ):
        # ie: "skip-pager" -> self.window._skip_pager
        if getattr(w, "_%s" % s.replace("-", "_"), False):
            state.append(s)
    for s in ("modal",):
        fn = getattr(w, "get_%s" % s, None)
        if fn and fn():
            state.append(s)
    return csv(state) or "none"


def get_window_attributes(w) -> str:
    attr = {}
    workspace = w.get_desktop_workspace()
    if workspace not in (None, WORKSPACE_UNSET):
        attr["workspace"] = workspace
    with IgnoreWarningsContext():
        opacity = w.get_opacity()
    if opacity < 1:
        attr["opacity"] = opacity
    role = w.get_role()
    if role:
        attr["role"] = role
    # get_type_hint
    return dict_str(attr)


class WindowInfo(Gtk.Window):

    def __init__(self, client, window):
        super().__init__()
        add_close_accel(self, self.close)
        self._client = client
        self._window = window
        self.is_closed = False
        self.set_title("Window Information for %s" % window.get_title())
        self.set_destroy_with_parent(True)
        self.set_resizable(True)
        self.set_decorated(True)
        self.set_transient_for(window)
        self.set_icon(get_icon_pixbuf("information.png"))
        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        def window_deleted(*_args) -> None:
            self.is_closed = True

        self.connect('delete_event', window_deleted)

        grid = Gtk.Grid()
        grid.set_row_homogeneous(False)
        grid.set_column_homogeneous(False)
        row = AtomicInteger()

        def new_row(text="", widget=None) -> None:
            lbl = label(text)
            lbl.set_xalign(1)
            lbl.set_margin_end(10)
            grid.attach(lbl, 1, int(row), 1, 1)
            if widget:
                grid.attach(widget, 2, int(row), 1, 1)
            row.increase()

        def lrow(text: str) -> Gtk.Label:
            lbl = label()
            lbl.set_margin_start(10)
            lbl.set_xalign(0)
            lbl.set_line_wrap(True)
            new_row(text, lbl)
            return lbl

        def irow(text) -> Gtk.Image:
            image = Gtk.Image()
            image.set_margin_start(10)
            image.set_halign(Gtk.Align.START)
            image.set_hexpand(False)
            new_row(text, image)
            return image

        def sep() -> None:
            s = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            s.set_margin_top(3)
            s.set_margin_bottom(3)
            grid.attach(s, 1, int(row), 2, 1)
            row.increase()

        self.wid_label = lrow("Window ID")
        self.title_label = lrow("Title")
        self.title_label.set_size_request(320, -1)
        self.rendering_label = lrow("Rendering")
        self.or_image = irow("Override-Redirect")
        self.state_label = lrow("State")
        self.attributes_label = lrow("Attributes")
        self.focus_image = irow("Focus")
        self.button_state_label = lrow("Button State")
        self.fps_label = lrow("Frames Per Second")
        sep()
        self.gravity_label = lrow("Gravity")
        self.content_type_label = lrow("Content Type")
        sep()
        self.pixel_depth_label = lrow("Pixel Depth")
        self.alpha_image = irow("Alpha Channel")
        self.opengl_image = irow("OpenGL")
        sep()
        self.geometry_label = lrow("Geometry")
        self.outer_geometry_label = lrow("Outer Geometry")
        self.inner_geometry_label = lrow("Inner Geometry")
        self.offsets_label = lrow("Offsets")
        self.frame_extents_label = lrow("Frame Extents")
        self.max_size_label = lrow("Maximum Size")
        self.size_constraints_label = lrow("Size Constraints")
        sep()
        # backing:
        self.video_properties = lrow("Video Decoder")
        sep()
        self.backing_properties = lrow("Backing Properties")
        sep()
        btn = Gtk.Button(label="Copy to clipboard")
        btn.connect("clicked", self.copy_to_clipboard)
        grid.attach(btn, 1, int(row), 2, 1)
        vbox = Gtk.VBox()
        vbox.pack_start(grid, True, True, 20)
        self.add(vbox)

    def close(self, *_args) -> None:
        self.is_closed = True
        self.hide()
        super().close()

    def show(self) -> None:
        self.populate()
        self.set_size_request(320, -1)
        super().show_all()
        GLib.timeout_add(1000, self.populate)

    def populate(self) -> bool:
        if self.is_closed:
            return False
        self.do_populate()
        return True

    def copy_to_clipboard(self, *_args) -> None:
        w = self._window
        if not w:
            return
        info = {
            "wid": w.wid,
            "title": w.get_title(),
            "override-redirect": w._override_redirect,
            "state": get_window_state(w),
            "attributes": get_window_attributes(w),
            "focused": w._focused,
            "buttons": csv(b for b, s in w.button_state.items() if s) or "none",
            "gravity": gravity_str(w.window_gravity),
            "content-type": w.content_type or "unknown",
            "pixel-depth": w.pixel_depth or 24,
            "alpha": w._window_alpha,
            "opengl": w.is_GL(),
            "geometry": geom_str(list(w._pos) + list(w._size)),
            "outer-geometry": geom_str(list(w.get_position()) + list(w.get_size())),
            "inner-geometry": geom_str(w.get_drawing_area_geometry()),
            "offsets": csv(str(x) for x in (w.window_offset or ())) or "none",
            "frame-extents": csv(w._current_frame_extents or []) or "none",
            "max-size": csv(w.max_window_size),
            "size-constraints": hsc(w.size_constraints),
        }
        # backing:
        b = w._backing
        if b:
            info |= {
                "size": csv(b.size),
                "render-size": csv(b.render_size),
                "backing-offsets": csv(b.offsets),
            }
        text = "\n".join(f"{k}={v}" for k, v in info.items())
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(text, len(text))

    def do_populate(self) -> None:
        w = self._window
        if not w:
            return
        fps = "n/a"
        b = w._backing
        binfo = {}
        if b:
            update_fps = getattr(b, "update_fps", None)
            if callable(update_fps):
                update_fps()
                fps = str(getattr(b, "fps_value", "n/a"))
            binfo = b.get_info()
        self.wid_label.set_text(str(w.wid))
        self.rendering_label.set_text(binfo.get("type", "unknown"))
        self.title_label.set_text(w.get_title())
        self.bool_icon(self.or_image, w._override_redirect)
        self.state_label.set_text(get_window_state(w))
        self.attributes_label.set_text(get_window_attributes(w))
        self.bool_icon(self.focus_image, w._focused)
        self.button_state_label.set_text(csv(b for b, s in w.button_state.items() if s) or "none")
        self.fps_label.set_text(fps)
        # self.group_leader_label.set_text(str(w.group_leader))
        self.gravity_label.set_text(gravity_str(w.window_gravity))
        self.content_type_label.set_text(w.content_type or "unknown")
        # geometry:
        self.pixel_depth_label.set_text(str(w.pixel_depth or 24))
        self.bool_icon(self.alpha_image, w._window_alpha)
        self.bool_icon(self.opengl_image, w.is_GL())
        # tells us if this window instance can paint with alpha
        geom = list(w._pos) + list(w._size)
        self.geometry_label.set_text(geom_str(geom))
        geom = list(w.get_position()) + list(w.get_size())
        self.outer_geometry_label.set_text(geom_str(geom))
        self.inner_geometry_label.set_text(geom_str(w.get_drawing_area_geometry()))
        self.offsets_label.set_text(csv(str(x) for x in (w.window_offset or ())) or "none")
        self.frame_extents_label.set_text(csv(w._current_frame_extents or []) or "none")
        self.max_size_label.set_text(csv(w.max_window_size))
        self.size_constraints_label.set_text(hsc(w.size_constraints))
        # backing:
        if b:
            self.backing_properties.show()

            def dict_to_str(d, sep="\n", eq="=", exclude=()) -> str:
                strdict = {k: pv(v) for k, v in d.items() if k not in exclude}
                return sep.join("%s%s%s" % (k, eq, v) for k, v in strdict.items() if v)

            def pv(value) -> str:
                if isinstance(value, (tuple, list)):
                    return csv(value)
                if isinstance(value, dict):
                    return dict_to_str(value, ", ", ":")
                return str(value)

            self.backing_properties.set_text(dict_to_str(binfo, exclude=(
                "transparency",
                "size",
                "render-size",
                "offsets",
                "fps",
                "mmap",
                "type",
                "bit-depth",
                "video-decoder",
            )))
            vdinfo = binfo.get("video-decoder")
            if vdinfo:
                self.video_properties.show()
                self.video_properties.set_text(dict_to_str(vdinfo))
            else:
                self.video_properties.hide()
        else:
            self.backing_properties.hide()
            self.backing_properties.set_text("")

    def bool_icon(self, image, on_off: bool) -> None:
        c = self._client
        if not c:
            return
        if on_off:
            icon = get_icon_pixbuf("ticked-small.png")
        else:
            icon = get_icon_pixbuf("unticked-small.png")
        image.set_from_pixbuf(icon)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.platform import platform_import
from xpra.log import Logger, consume_verbose_argv

log = Logger("webcam")


def get_virtual_video_devices() -> dict:
    return {}


def get_all_video_devices():
    # None means we can't enumerate,
    # this is different from an empty dict!
    return None


def add_video_device_change_callback(_callback: Callable) -> None:
    # not implemented here
    pass


def remove_video_device_change_callback(_callback: Callable) -> None:
    # not implemented here
    pass


_video_device_change_callbacks: list[Callable] = []


def _fire_video_device_change(create=None, pathname=None) -> None:
    for callback in _video_device_change_callbacks:
        try:
            callback(create, pathname)
        except Exception as e:
            log("error on %s", callback, exc_info=True)
            log.error("Error: video device change callback error")
            log.estr(e)


platform_import(globals(), "webcam", False,
                "get_virtual_video_devices",
                "get_all_video_devices",
                "add_video_device_change_callback",
                "remove_video_device_change_callback")


def main(argv) -> int:
    run = "-r" in argv or "--run" in argv
    from xpra.util.str_fn import print_nested_dict
    from xpra.platform import program_context
    with program_context("Webcam Info", "Webcam Info"):
        consume_verbose_argv(argv, "webcam")
        devices = get_virtual_video_devices() or {}
        log.info("Found %i virtual video devices:", len(devices))
        print_nested_dict(devices)
        all_devices = get_all_video_devices() or {}
        log.info("Found %i video devices in total:", len(all_devices))
        print_nested_dict(all_devices)

        if run:
            log.info("add watch for video device changes")

            def callback(added=None, device=None):
                if added is not None or device:
                    log.info("video device %s: %s", ["removed", "added"][added], device)
                else:
                    log.info("device change")

            log.info("starting main loop")
            glib = gi_import("GLib")
            main_loop = glib.MainLoop()
            glib.idle_add(add_video_device_change_callback, callback)
            try:
                main_loop.run()
            except KeyboardInterrupt:
                pass
            log.info("terminating, removing callback")
            remove_video_device_change_callback(callback)
    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.os_util import (
    POSIX,
    getuid, getgid, get_shell_for_uid, get_username_for_uid, get_home_for_uid,
    get_hex_uuid, get_int_uuid, get_user_uuid,
)
from xpra.util.env import OSEnvContext
from xpra.util.thread import is_main_thread
from xpra.util.io import livefds
from xpra.util.system import (
    is_Ubuntu, is_Debian, is_unity,
    is_gnome, is_kde, is_WSL,
)
from xpra.util.str_fn import strtobytes, bytestostr, hexstr, memoryview_to_bytes


class TestOSUtil(unittest.TestCase):

    def check(self, str_value):
        b = strtobytes(str_value)
        assert b
        s = bytestostr(b)
        assert s
        assert s==str_value
        if not memoryview:
            return
        mv = memoryview(b)
        mvb = memoryview_to_bytes(mv)
        mvs = bytestostr(mvb)
        assert mvs==str_value

    def test_livefds(self):
        assert len(livefds())>=2

    def test_distribution_variant(self):
        is_Ubuntu()
        is_Debian()
        is_WSL()

    def test_de(self):
        is_unity()
        is_gnome()
        is_kde()

    def test_uuid(self):
        assert len(get_hex_uuid())==32
        assert isinstance(get_int_uuid(), int)
        assert get_int_uuid()!=0
        assert get_user_uuid()!=0

    def test_posix_wrappers(self):
        if not POSIX:
            return
        assert isinstance(getuid(), int)
        assert isinstance(getgid(), int)
        def isstr(v):
            assert v
            assert isinstance(v, str)
        isstr(get_shell_for_uid(getuid()))
        isstr(get_username_for_uid(getuid()))
        isstr(get_home_for_uid(getuid()))
        assert not get_shell_for_uid(999999999)


    def test_memoryview_to_bytes(self):
        assert memoryview_to_bytes(b"bar")==b"bar"
        assert memoryview_to_bytes(memoryview(b"foo"))==b"foo"
        assert memoryview_to_bytes(bytearray(b"foo"))==b"foo"
        assert memoryview_to_bytes(u"foo")==b"foo"

    def test_hexstr(self):
        assert hexstr("01")=="3031"

    def test_bytes(self):
        assert strtobytes(b"hello")==b"hello"

    def test_strs(self):
        assert bytestostr(u"foo")==u"foo"
        for l in (1, 16, 255):
            zeroes  = chr(0)*l
            ones    = chr(1)*l
            ff      = chr(255)*l
            fe      = chr(254)*l
            self.check(zeroes)
            self.check(ones)
            self.check(ff)
            self.check(fe)

    def test_env_context(self):
        import os
        env = os.environ.copy()
        with OSEnvContext():
            os.environ["foo"] = "bar"
        assert os.environ.get("foo")!="bar"
        assert os.environ==env

    def test_is_main_thread(self):
        assert is_main_thread()
        result = []
        def notmainthread():
            result.append(is_main_thread())
        from threading import Thread
        t = Thread(target=notmainthread)
        t.start()
        t.join()
        assert len(result)==1
        assert result[0] is False


def main():
    unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from io import BytesIO
from typing import Any
from collections.abc import Sequence, Callable

from xpra.os_util import gi_import
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.net.compression import Compressed
from xpra.util.str_fn import memoryview_to_bytes
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("server", "cursor")

SAVE_CURSORS = envbool("XPRA_SAVE_CURSORS", False)


class CursorsMixin(StubSourceMixin):
    PREFIX = "cursor"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        return caps.boolget("cursors")

    def __init__(self):
        self.get_cursor_data_cb: Callable | None = None
        self.send_cursors = False
        self.cursor_encodings: Sequence[str] = ()
        self.cursor_timer = 0
        self.last_cursor_sent: tuple = ()

    def init_from(self, _protocol, server) -> None:
        self.get_cursor_data_cb = server.get_cursor_data

    def init_state(self) -> None:
        # WindowSource for each Window ID
        self.send_cursors = False
        self.cursor_encodings = ()
        self.last_cursor_sent = ()

    def cleanup(self) -> None:
        self.cancel_cursor_timer()

    def resume(self) -> None:
        self.send_cursor()

    def parse_client_caps(self, c: typedict) -> None:
        self.send_cursors = self.send_windows and c.boolget("cursors")
        self.cursor_encodings = c.strtupleget("encodings.cursor")
        log(f"cursors={self.send_cursors}, cursor encodings={self.cursor_encodings}")

    def get_caps(self) -> dict[str, Any]:
        return {}

    ######################################################################
    # info:
    def get_info(self) -> dict[str, Any]:
        return {
            CursorsMixin.PREFIX: {
                "enabled": self.send_cursors,
                "encodings": self.cursor_encodings,
            },
        }

    def send_cursor(self) -> None:
        if not self.send_cursors or self.suspended or not self.hello_sent:
            return
        # if not pending already, schedule it:
        gbc = self.global_batch_config
        if not self.cursor_timer and gbc:
            delay = max(10, int(gbc.delay / 4))

            def do_send_cursor():
                self.cursor_timer = 0
                cd = self.get_cursor_data_cb()
                if not cd or not cd[0]:
                    self.send_empty_cursor()
                    return
                cursor_data = list(cd[0])
                cursor_sizes = cd[1]
                self.do_send_cursor(delay, cursor_data, cursor_sizes)

            self.cursor_timer = GLib.timeout_add(delay, do_send_cursor)

    def cancel_cursor_timer(self) -> None:
        ct = self.cursor_timer
        if ct:
            self.cursor_timer = 0
            GLib.source_remove(ct)

    def do_send_cursor(self, delay, cursor_data, cursor_sizes, encoding_prefix="") -> None:
        # x11 server core calls this method directly, so check availability again:
        if not self.send_cursors:
            return
        # copy to a new list we can modify (ie: compress):
        cursor_data = list(cursor_data)
        # skip first two fields (if present) as those are coordinates:
        if self.last_cursor_sent and self.last_cursor_sent[2:9] == cursor_data[2:9]:
            log("do_send_cursor(..) cursor identical to the last one we sent, nothing to do")
            return
        self.last_cursor_sent = cursor_data[:9]
        w, h, _xhot, _yhot, serial, pixels, name = cursor_data[2:9]
        # compress pixels if needed:
        encoding = "raw"
        if pixels is not None:
            cpixels: bytes | Compressed = memoryview_to_bytes(pixels)
            try:
                from PIL import Image
            except ImportError:
                Image = None
            if "png" in self.cursor_encodings and Image:
                log(f"do_send_cursor() got {len(cpixels)} bytes of pixel data for {w}x{h} cursor named {name!r}")
                img = Image.frombytes("RGBA", (w, h), cpixels, "raw", "BGRA", w * 4, 1)
                buf = BytesIO()
                img.save(buf, "PNG")
                pngdata = buf.getvalue()
                buf.close()
                cpixels = Compressed("png cursor", pngdata)
                encoding = "png"
                if SAVE_CURSORS:
                    filename = f"raw-cursor-{serial:x}.png"
                    with open(filename, "wb") as f:
                        f.write(pngdata)
                    log("cursor saved to %s", filename)
            elif len(cpixels) >= 256 and ("raw" in self.cursor_encodings or not self.cursor_encodings):
                cpixels = self.compressed_wrapper("cursor", pixels)
                log("do_send_cursor(..) pixels=%s ", cpixels)
                encoding = "raw"
            else:
                log("no supported cursor encodings")
                return
            cursor_data[7] = cpixels
        log("do_send_cursor(..) %sx%s %s cursor name='%s', serial=%#x with delay=%s (cursor_encodings=%s)",
            w, h, (encoding or "empty"), name, serial, delay, self.cursor_encodings)
        args = [encoding_prefix + encoding] + list(cursor_data[:9]) + [cursor_sizes[0]] + list(cursor_sizes[1])
        self.send_more("cursor", *args)

    def send_empty_cursor(self) -> None:
        log("send_empty_cursor(..)")
        self.last_cursor_sent = ()
        self.send_more("cursor", "")



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.window import add_close_accel

Gtk = gi_import("Gtk")
Gio = gi_import("Gio")
GLib = gi_import("GLib")


class HeaderBarWindow(Gtk.Window):

    def __init__(self):
        Gtk.Window.__init__(self, title="Stack Demo")
        self.set_border_width(10)
        self.set_default_size(400, 200)
        self.set_position(Gtk.WindowPosition.CENTER)

        hb = Gtk.HeaderBar()
        hb.set_show_close_button(True)
        hb.props.title = "HeaderBar example"
        self.set_titlebar(hb)

        button = Gtk.Button()
        icon = Gio.ThemedIcon(name="mail-send-receive-symbolic")
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        button.add(image)
        hb.pack_end(button)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        Gtk.StyleContext.add_class(box.get_style_context(), "linked")

        button = Gtk.Button()
        button.add(Gtk.Arrow(arrow_type=Gtk.ArrowType.LEFT, shadow_type=Gtk.ShadowType.NONE))
        box.add(button)

        button = Gtk.Button()
        button.add(Gtk.Arrow(arrow_type=Gtk.ArrowType.RIGHT, shadow_type=Gtk.ShadowType.NONE))
        box.add(button)

        hb.pack_start(box)

        self.add(Gtk.TextView())

    def show_with_focus(self):
        force_focus()
        self.show_all()
        super().present()


def main():
    with program_context("header-bar", "Header Bar"):
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("header bar test window")
        w = HeaderBarWindow()
        w.connect("delete-event", Gtk.main_quit)
        add_close_accel(w, Gtk.main_quit)
        GLib.idle_add(w.show_with_focus)
        Gtk.main()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# cython code for manipulating GdkAtoms
from typing import List

from xpra.os_util import gi_import
from libc.stdint cimport uintptr_t   # pylint: disable=syntax-error

Gdk = gi_import("Gdk")


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "gtk-3.0/gdk/gdk.h":
    ctypedef void* GdkAtom
    GdkAtom GDK_NONE

cdef extern from "gtk-3.0/gdk/gdkproperty.h":
    ctypedef char gchar
    ctypedef int gint
    ctypedef gint gboolean
    gchar* gdk_atom_name(GdkAtom atom)
    GdkAtom gdk_atom_intern(const gchar *atom_name, gboolean only_if_exists)


def gdk_atom_objects_from_gdk_atom_array(atom_string) -> List[Gdk.Atom]:
    cdef Py_buffer py_buf
    if PyObject_GetBuffer(atom_string, &py_buf, PyBUF_ANY_CONTIGUOUS):
        raise ValueError(f"failed to read atom buffer of {type(atom_string)}")
    cdef unsigned int array_len = py_buf.len // sizeof(GdkAtom)
    objects = []
    cdef unsigned int i
    cdef const GdkAtom * array = <GdkAtom*> py_buf.buf
    cdef GdkAtom atom
    cdef gchar*  name
    for i in range(array_len):
        atom = array[i]
        if atom==GDK_NONE:
            continue
        #round-trip via the name,
        #inefficient but what other constructor is there?
        name = gdk_atom_name(atom)
        if name:
            str_name = name.decode("latin1")
            gdk_atom = Gdk.Atom.intern(str_name, False)
            objects.append(gdk_atom)
    PyBuffer_Release(&py_buf)
    return objects


def gdk_atom_array_from_atoms(atoms) -> List[int]:
    cdef GdkAtom gdk_atom
    cdef uintptr_t gdk_atom_value
    atom_array = []
    for atom in atoms:
        gdk_atom = gdk_atom_intern(atom, False)
        if gdk_atom!=GDK_NONE:
            gdk_atom_value = <uintptr_t> gdk_atom
            atom_array.append(gdk_atom_value)
    return atom_array



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys


def main(argv=()) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("Xpra-Bug-Report", "Xpra Bug Report"):
        from xpra.gtk.dialogs import bug_report
        return bug_report.main(argv)


if __name__ == "__main__":
    v = main(sys.argv)
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from typing import Any

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.server.source.windows import WindowsMixin
from xpra.net.common import PacketType
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("window")
focuslog = Logger("focus")
metalog = Logger("metadata")
geomlog = Logger("geometry")
eventslog = Logger("events")


class WindowServer(StubServerMixin):
    """
    Mixin for servers that forward windows.
    """
    PREFIX = "window"

    def __init__(self):
        # Window id 0 is reserved for "not a window"
        self._max_window_id = 1
        self._window_to_id: dict[Any, int] = {}
        self._id_to_window: dict[int, Any] = {}
        self.window_filters = []
        self.window_min_size = 0, 0
        self.window_max_size = 2 ** 15 - 1, 2 ** 15 - 1

    def init(self, opts) -> None:
        def parse_window_size(v, default_value=(0, 0)):
            try:
                # split on "," or "x":
                pv = tuple(int(x.strip()) for x in v.replace(",", "x").split("x", 1))
                assert len(pv) == 2
                w, h = pv
                if w <= 0 or h <= 0 or w >= 32768 or h >= 32768:
                    raise ValueError(f"invalid window size {w}x{h}")
                return w, h
            except Exception:
                return default_value

        self.window_min_size = parse_window_size(opts.min_size, (0, 0))
        self.window_max_size = parse_window_size(opts.max_size, (2 ** 15 - 1, 2 ** 15 - 1))
        minw, minh = self.window_min_size
        maxw, maxh = self.window_max_size
        self.update_size_constraints(minw, minh, maxw, maxh)

    def setup(self) -> None:
        self.load_existing_windows()
        self.add_init_thread_callback(self.reinit_window_encoders)

    def cleanup(self) -> None:
        for window in tuple(self._window_to_id.keys()):
            window.unmanage()
        # this can cause errors if we receive packets during shutdown:
        # self._window_to_id = {}
        # self._id_to_window = {}

    def reinit_window_encoders(self) -> None:
        # any window mapped before the threaded init completed
        # may need to re-initialize its list of encoders:
        log("reinit_window_encoders()")
        for ss in self._server_sources.values():
            if isinstance(ss, WindowsMixin):
                ss.reinit_encoders()

    def last_client_exited(self) -> None:
        self._focus(None, 0, [])

    def get_server_features(self, _source) -> dict[str, Any]:
        return {}

    def get_info(self, _proto) -> dict[str, Any]:
        return {
            "state": {
                "windows": sum(int(window.is_managed()) for window in tuple(self._id_to_window.values())),
            },
            "filters": tuple((uuid, repr(f)) for uuid, f in self.window_filters),
        }

    def get_ui_info(self, _proto, _client_uuids=None, wids=None, *_args) -> dict[str, Any]:
        """ info that must be collected from the UI thread
            (ie: things that query the display)
        """
        return {"windows": self.get_windows_info(wids)}

    def parse_hello(self, ss, caps: typedict, send_ui) -> None:
        if send_ui:
            self.parse_hello_ui_window_settings(ss, caps)

    def parse_hello_ui_window_settings(self, ss, c: typedict) -> None:
        """
        this method is overriden by the `seamless` server to set frame extents
        """

    def add_new_client(self, *_args) -> None:
        minw, minh = self.window_min_size
        maxw, maxh = self.window_max_size
        for ss in tuple(self._server_sources.values()):
            if not isinstance(ss, WindowsMixin):
                continue
            cminw, cminh = ss.window_min_size
            cmaxw, cmaxh = ss.window_max_size
            minw = max(minw, cminw)
            minh = max(minh, cminh)
            if cmaxw > 0:
                maxw = min(maxw, cmaxw)
            if cmaxh > 0:
                maxh = min(maxh, cmaxh)
        maxw = max(1, maxw)
        maxh = max(1, maxh)
        if minw > 0 and minw > maxw:
            maxw = minw
        if minh > 0 and minh > maxh:
            maxh = minh
        self.update_size_constraints(minw, minh, maxw, maxh)

    def update_size_constraints(self, minw: int, minh: int, maxw: int, maxh: int) -> None:
        # subclasses may update the window models
        pass

    def send_initial_data(self, ss, caps, send_ui: bool, share_count: int) -> None:
        if not send_ui:
            return
        if not isinstance(ss, WindowsMixin):
            return
        self.send_initial_windows(ss, share_count > 0)

    def is_shown(self, _window) -> bool:
        return True

    def get_window_id(self, _gdkwindow) -> int:
        # the X11 server can find the model from the window's xid
        return 0

    def reset_window_filters(self) -> None:
        self.window_filters = []

    def get_windows_info(self, window_ids) -> dict[int, dict[str, Any]]:
        info = {}
        for wid, window in self._id_to_window.items():
            if window_ids is not None and wid not in window_ids:
                continue
            info[wid] = self.get_window_info(window)
        return info

    def get_window_info(self, window) -> dict[str, Any]:
        from xpra.server.window.metadata import make_window_metadata
        info = {}
        for prop in window.get_property_names():
            if prop == "icons" or prop is None:
                continue
            metadata = make_window_metadata(window, prop)
            info.update(metadata)
        for prop in window.get_internal_property_names():
            metadata = make_window_metadata(window, prop)
            info.update(metadata)
        info.update({
            "override-redirect": window.is_OR(),
            "tray": window.is_tray(),
            "size": window.get_dimensions(),
        })
        wid = self._window_to_id.get(window)
        if wid:
            wprops = self.client_properties.get(wid)
            if wprops:
                info["client-properties"] = wprops
        return info

    def _update_metadata(self, window, pspec) -> None:
        metalog("updating metadata on %s: %s", window, pspec)
        wid = self._window_to_id.get(window)
        if not wid:
            return  # window is already gone
        for ss in self._server_sources.values():
            if isinstance(ss, WindowsMixin):
                ss.window_metadata(wid, window, pspec.name)

    def _remove_window(self, window) -> int:
        wid = self._window_to_id[window]
        log("remove_window: %s - %s", wid, window)
        for ss in self._server_sources.values():
            if isinstance(ss, WindowsMixin):
                ss.lost_window(wid, window)
        del self._window_to_id[window]
        del self._id_to_window[wid]
        for ss in self._server_sources.values():
            if isinstance(ss, WindowsMixin):
                ss.remove_window(wid, window)
        self.client_properties.pop(wid, None)
        return wid

    def _add_new_window_common(self, window) -> int:
        wid = self._max_window_id
        self.do_add_new_window_common(wid, window)
        return wid

    def do_add_new_window_common(self, wid: int, window) -> None:
        self._max_window_id = max(self._max_window_id, wid + 1)
        props = window.get_dynamic_property_names()
        metalog("add_new_window_common(%s) watching for dynamic properties: %s", window, props)
        for prop in props:
            window.managed_connect("notify::%s" % prop, self._update_metadata)
        self._window_to_id[window] = wid
        self._id_to_window[wid] = window

    def _do_send_new_window_packet(self, ptype, window, geometry) -> None:
        wid = self._window_to_id[window]
        for ss in tuple(self._server_sources.values()):
            if not isinstance(ss, WindowsMixin):
                continue
            wprops = self.client_properties.get(wid, {}).get(ss.uuid, {})
            x, y, w, h = geometry
            # adjust if the transient-for window is not mapped in the same place by the client we send to:
            if "transient-for" in window.get_property_names():
                transient_for = window.get_property("transient-for")
                if transient_for:
                    window_tf = self.get_window_id(transient_for)
                    parent = self._id_to_window.get(window_tf)
                    parent_ws = ss.get_window_source(window_tf)
                    pos = self.get_window_position(parent)
                    geomlog("transient-for=%s : %s, ws=%s, pos=%s", transient_for, parent, parent_ws, pos)
                    if pos and parent and parent_ws:
                        mapped_at = parent_ws.mapped_at
                        if mapped_at:
                            wx, wy = pos
                            cx, cy = mapped_at[:2]
                            if wx != cx or wy != cy:
                                dx, dy = cx - wx, cy - wy
                                if dx or dy:
                                    geomlog("adjusting new window position for client window offset: %s", (dx, dy))
                                    x += dx
                                    y += dy
            ss.new_window(ptype, wid, window, x, y, w, h, wprops)

    def _process_damage_sequence(self, proto, packet: PacketType) -> None:
        packet_sequence, wid, width, height, decode_time = packet[1:6]
        if len(packet) >= 7:
            message = packet[6]
        else:
            message = ""
        ss = self.get_server_source(proto)
        if ss:
            ss.client_ack_damage(packet_sequence, wid, width, height, decode_time, message)

    def refresh_window(self, window) -> None:
        ww, wh = window.get_dimensions()
        self.refresh_window_area(window, 0, 0, ww, wh)

    def refresh_window_area(self, window, x, y, width, height, options=None) -> None:
        wid = self._window_to_id[window]
        for ss in tuple(self._server_sources.values()):
            damage = getattr(ss, "damage", None)
            if damage:
                damage(wid, window, x, y, width, height, options)

    def _process_buffer_refresh(self, proto, packet: PacketType) -> None:
        """ can be used for requesting a refresh, or tuning batch config, or both """
        wid, _, qual = packet[1:4]
        if len(packet) >= 6:
            options = typedict(packet[4])
            client_properties = packet[5]
        else:
            options = typedict({})
            client_properties = {}
        if wid == -1:
            wid_windows = self._id_to_window
        elif wid in self._id_to_window:
            wid_windows = {wid: self._id_to_window.get(wid)}
        else:
            # may have been destroyed since the request was made
            log("invalid window specified for refresh: %s", wid)
            return
        log("process_buffer_refresh for windows: %s options=%s, client_properties=%s",
            wid_windows, options, client_properties)
        batch_props = options.dictget("batch", {})
        if batch_props or client_properties:
            # change batch config and/or client properties
            self.update_batch_config(proto, wid_windows, typedict(batch_props), client_properties)
        # default to True for backwards compatibility:
        if options.get("refresh-now", True):
            refresh_opts = {"quality": qual,
                            "override_options": True}
            self._refresh_windows(proto, wid_windows, refresh_opts)

    def update_batch_config(self, proto, wid_windows, batch_props, client_properties) -> None:
        ss = self.get_server_source(proto)
        if ss is None:
            return
        for wid, window in wid_windows.items():
            if window is None or not window.is_managed():
                continue
            self._set_client_properties(proto, wid, window, client_properties)
            ss.update_batch(wid, window, batch_props)

    def _refresh_windows(self, proto, wid_windows, opts=None) -> None:
        ss = self.get_server_source(proto)
        if ss:
            self.do_refresh_windows(ss, wid_windows, opts)

    def do_refresh_windows(self, ss, wid_windows, opts=None) -> None:
        for wid, window in wid_windows.items():
            if window is None or not window.is_managed():
                continue
            if not self.is_shown(window):
                log("window is no longer shown, ignoring buffer refresh which would fail")
                continue
            ss.refresh(wid, window, opts)

    def _idle_refresh_all_windows(self, proto) -> None:
        GLib.idle_add(self._refresh_windows, proto, self._id_to_window, {})

    def refresh_all_windows(self) -> None:
        for ss in tuple(self._server_sources.values()):
            if not isinstance(ss, WindowsMixin):
                self.do_refresh_windows(ss, self._id_to_window)

    def get_window_position(self, _window) -> tuple[int, int] | None:
        # where the window is actually mapped on the server screen:
        return None

    def _window_mapped_at(self, proto, wid: int, window, coords=None) -> None:
        # record where a window is mapped by a client
        # (in order to support multiple clients and different offsets)
        ss = self.get_server_source(proto)
        if not ss:
            return
        if coords:
            ss.map_window(wid, window, coords)
        else:
            ss.unmap_window(wid, window)

    def _process_map_window(self, proto, packet: PacketType) -> None:
        log.info("_process_map_window(%s, %s)", proto, packet)

    def _process_unmap_window(self, proto, packet: PacketType) -> None:
        log.info("_process_unmap_window(%s, %s)", proto, packet)

    def _process_close_window(self, proto, packet: PacketType) -> None:
        log.info("_process_close_window(%s, %s)", proto, packet)

    def _process_configure_window(self, proto, packet: PacketType) -> None:
        log.info("_process_configure_window(%s, %s)", proto, packet)

    def send_initial_windows(self, ss, sharing=False) -> None:
        raise NotImplementedError()

    ######################################################################
    # focus:
    def _process_focus(self, proto, packet: PacketType) -> None:
        if self.readonly:
            return
        wid = packet[1]
        focuslog("process_focus: wid=%s", wid)
        if len(packet) >= 3:
            modifiers = packet[2]
        else:
            modifiers = None
        ss = self.get_server_source(proto)
        if ss:
            self._focus(ss, wid, modifiers)
            # if the client focused one of our windows, count this as a user event:
            if wid > 0:
                ss.user_event()

    def _focus(self, _server_source, wid, modifiers) -> None:
        focuslog("_focus(%s,%s)", wid, modifiers)

    def get_focus(self) -> int:
        # can be overridden by subclasses that do manage focus
        # (ie: not shadow servers which only have a single window)
        # default: no focus
        return -1

    def init_packet_handlers(self) -> None:
        self.add_packets(
            "map-window", "unmap-window",
            "configure-window", "close-window",
            "focus",
            "damage-sequence", "buffer-refresh",
            main_thread=True)



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# Platform-specific code for Win32.
# pylint: disable=bare-except
# pylint: disable=import-outside-toplevel

import sys
import errno
import os.path
import ctypes
from ctypes import WINFUNCTYPE, WinDLL, POINTER, byref, wintypes, create_unicode_buffer  # @UnresolvedImport
from ctypes.wintypes import BOOL, HANDLE, DWORD, LPWSTR, LPVOID, WORD, SMALL_RECT

from xpra.util.env import envbool
from xpra.common import noerr
from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.common import (
    GetStdHandle,
    SetConsoleTitleA, GetConsoleScreenBufferInfo,
    MessageBoxA, GetLastError, kernel32,
)


if not sys.platform.startswith("win"):
    raise ImportError("not to be used on %s" % sys.platform)

STD_INPUT_HANDLE = DWORD(-10)
STD_OUTPUT_HANDLE = DWORD(-11)
STD_ERROR_HANDLE = DWORD(-12)
GetFileType = WINFUNCTYPE(DWORD, DWORD)(("GetFileType", kernel32))
FILE_TYPE_CHAR = 0x0002
FILE_TYPE_REMOTE = 0x8000
GetConsoleMode = WINFUNCTYPE(BOOL, HANDLE, POINTER(DWORD))(("GetConsoleMode", kernel32))
INVALID_HANDLE_VALUE = DWORD(-1).value
STDOUT_FILENO = 1
STDERR_FILENO = 2
WriteConsoleW = WINFUNCTYPE(BOOL, HANDLE, LPWSTR, DWORD, POINTER(DWORD), LPVOID)(("WriteConsoleW", kernel32))

CSIDL_COMMON_STARTMENU = 0x16
CSIDL_STARTMENU = 0xb
CSIDL_COMMON_APPDATA = 35

GetConsoleCP = kernel32.GetConsoleCP  # @UndefinedVariable

# redirect output if we're not running from a console:
FROZEN: bool = getattr(sys, "frozen", None) in ("windows_exe", "console_exe", True)
REDIRECT_OUTPUT: bool = envbool("XPRA_REDIRECT_OUTPUT", FROZEN and GetConsoleCP() == 0)


def is_wine() -> bool:
    try:
        import winreg
        h_key = winreg.OpenKey(win32con.HKEY_LOCAL_MACHINE, r"Software\\Wine")
        return bool(h_key)
    except OSError:
        # no wine key, assume not present and wait for input
        pass
    return False


def get_csidl_folder(csidl: int) -> str:
    try:
        buf = create_unicode_buffer(wintypes.MAX_PATH)
        shell32 = WinDLL("shell32", use_last_error=True)
        get_folder_path = shell32.SHGetFolderPathW
        get_folder_path(0, csidl, None, 0, buf)
        return os.path.normpath(buf.value)
    except OSError:
        return ""


def get_common_startmenu_dir() -> str:
    return get_csidl_folder(CSIDL_COMMON_STARTMENU)


def get_startmenu_dir() -> str:
    return get_csidl_folder(CSIDL_STARTMENU)


def get_commonappdata_dir() -> str:
    return get_csidl_folder(CSIDL_COMMON_APPDATA)


prg_name = "Xpra"


def set_prgname(name: str) -> None:
    global prg_name
    prg_name = name
    try:
        SetConsoleTitleA(name.encode("latin1"))
    except OSError:
        pass


def not_a_console(handle: int) -> bool:
    if handle == INVALID_HANDLE_VALUE or handle is None:
        return True
    if (GetFileType(handle) & ~FILE_TYPE_REMOTE) != FILE_TYPE_CHAR:
        return True
    return GetConsoleMode(handle, byref(DWORD())) == 0


class COORD(ctypes.Structure):
    _fields_ = [
        ('X', ctypes.c_short),
        ('Y', ctypes.c_short)
    ]


class CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
    _fields_ = [
        ("dwSize", COORD),
        ("dwCursorPosition", COORD),
        ("wAttributes", WORD),
        ("srWindow", SMALL_RECT),
        ("dwMaximumWindowSize", COORD),
    ]


def get_console_position() -> tuple[int, int]:
    try:
        handle = GetStdHandle(STD_OUTPUT_HANDLE)
        if not_a_console(handle):
            return -1, -1
        # handle.SetConsoleTextAttribute(FOREGROUND_BLUE)
        csbi = CONSOLE_SCREEN_BUFFER_INFO()
        GetConsoleScreenBufferInfo(handle, byref(csbi))
        cpos = csbi.dwCursorPosition
        # wait for input if this is a brand-new console:
        return cpos.X, cpos.Y
    except Exception:
        e = sys.exc_info()[1]
        code = -1
        try:
            code = e.winerror
        except AttributeError:
            pass
        if code == errno.ENXIO:
            # ignore "no device" errors silently
            # (ie: happens if you redirect the command to a file)
            # we could also re-use the code above from "not_a_console()"
            pass
        else:
            from xpra.log import Logger
            log = Logger("win32")
            try:
                log.error("Error accessing console %s: %s" % (
                    errno.errorcode.get(e.errno, e.errno), e))
            except Exception:
                log.error(f"Error accessing console: {e}")
        return -1, -1


initial_console_position = get_console_position()
_wait_for_input = False


def should_wait_for_input() -> bool:
    wfi = os.environ.get("XPRA_WAIT_FOR_INPUT")
    if wfi is not None:
        return wfi != "0"
    if is_wine():
        # don't wait for input when running under wine
        # (which usually does not pop up a new shell window)
        return False
    if os.environ.get("TERM", "") == "xterm":
        # msys, cygwin and git bash environments don't pop up a new shell window,
        # and they all set TERM=xterm
        return False
    handle = GetStdHandle(STD_OUTPUT_HANDLE)
    if not_a_console(handle):
        return False
    # wait for input if this is a brand-new console:
    return initial_console_position == (0, 0)


def is_terminal() -> bool:
    if os.environ.get("TERM", "") == "xterm":
        return True
    handle = GetStdHandle(STD_OUTPUT_HANDLE)
    if not_a_console(handle):
        return False
    # wait for input if this is a brand-new console:
    return initial_console_position != (-1, -1)


def setup_console_event_listener(handler, enable: bool) -> bool:
    from xpra.log import Logger
    log = Logger("win32")
    try:
        from xpra.platform.win32.common import SetConsoleCtrlHandler, ConsoleCtrlHandler
        log("calling SetConsoleCtrlHandler(%s, %s)", handler, enable)
        ctypes_handler = ConsoleCtrlHandler(handler)
        result = SetConsoleCtrlHandler(ctypes_handler, enable)
        log("SetConsoleCtrlHandler(%s, %s)=%s", handler, enable, result)
        if result == 0:
            log.error("Error: could not %s console control handler:", "set" if enable else "unset")
            log.error(" SetConsoleCtrlHandler: %r", GetLastError())
            return False
        return True
    except Exception as e:
        log.error("SetConsoleCtrlHandler error: %s", e)
        return False


def do_init() -> None:
    def datadir() -> str:
        from xpra.platform.win32.paths import get_appdata_dir
        appdatadir = get_appdata_dir(False)
        if not os.path.exists(appdatadir):
            os.mkdir(appdatadir)
        return appdatadir

    if FROZEN:
        if envbool("PYTHONDONTWRITEBYTECODE", False):
            sys.dont_write_bytecode = True
        if not os.environ.get("PYTHONPYCACHEPREFIX"):
            sys.pycache_prefix = os.path.join(datadir(), "pycache-%i.%i" % (sys.version_info[0], sys.version_info[1]))

    global _wait_for_input

    if not REDIRECT_OUTPUT or is_terminal():
        # figure out if we want to wait for input at the end:
        _wait_for_input = should_wait_for_input()
        return

    if envbool("XPRA_LOG_TO_FILE", True):
        log_filename = os.environ.get("XPRA_LOG_FILENAME")
        if not log_filename:
            from xpra.platform import get_prgname
            log_filename = (get_prgname() or "Xpra") + ".log"
        if not os.path.isabs(log_filename):
            log_filename = os.path.join(datadir(), log_filename)
        sys.stdout = open(log_filename, "a", encoding="utf8")
        sys.stderr = sys.stdout
        os.environ["XPRA_LOG_FILENAME"] = log_filename
        _wait_for_input = False


def do_init_env() -> None:
    from xpra.platform import init_env_common
    init_env_common()
    if os.environ.get("CRYPTOGRAPHY_OPENSSL_NO_LEGACY") is None:
        os.environ["CRYPTOGRAPHY_OPENSSL_NO_LEGACY"] = "1"
    # if os.environ.get("GDK_WIN32_DISABLE_HIDPI") is None:
    #    os.environ["GDK_WIN32_DISABLE_HIDPI"] = "1"
    #    os.environ["GDK_WIN32_PER_MONITOR_HIDPI"] = "1"
    if FROZEN:
        # cx_freeze paths:
        PATH = os.environ.get("PATH", "").split(os.pathsep)
        edir = os.path.abspath(os.path.dirname(sys.executable))
        libdir = os.path.join(edir, "lib")
        for d in (libdir, edir):
            if not os.path.exists(d) or not os.path.isdir(d):
                continue
            try:
                sys.path.remove(d)
            except ValueError:
                pass
            sys.path.insert(0, d)
            try:
                PATH.remove(d)
            except ValueError:
                pass
            PATH.insert(0, d)
        # CUDA:
        if not os.environ.get("CUDA_PATH"):
            os.environ["CUDA_PATH"] = edir
        # Gtk and gi:
        os.environ['GI_TYPELIB_PATH'] = os.path.join(libdir, "girepository-1.0")
        os.environ["PATH"] = os.pathsep.join(PATH)
        if not os.environ.get("GTK_THEME") and not os.environ.get("GTK_DEBUG"):
            for theme in ("Windows-10", "win32"):
                tdir = os.path.join(edir, "share", "themes", theme)
                if os.path.exists(tdir):
                    os.environ["GTK_THEME"] = theme
                    break
        # GStreamer's plugins:
        gst_dir = os.path.join(libdir, "gstreamer-1.0")   # ie: C:\Program Files\Xpra\lib\gstreamer-1.0
        os.environ["GST_PLUGIN_PATH"] = gst_dir


MB_ICONEXCLAMATION = 0x00000030
MB_ICONINFORMATION = 0x00000040
MB_SYSTEMMODAL = 0x00001000


def _show_message(message: str, utype) -> None:
    # TODO: detect cx_freeze equivalent
    show_messagebox = envbool("XPRA_MESSAGEBOX", REDIRECT_OUTPUT)
    from xpra.log import Logger
    log = Logger("win32")
    if show_messagebox:
        # try to use an alert box since no console output will be shown:
        try:
            MessageBoxA(0, message.encode(), prg_name.encode(), utype)
            return
        except Exception as e:
            log.error(f"Error: cannot show alert box: {e}")
    log.info(message)


def command_info(message) -> None:
    _show_message(message, MB_ICONINFORMATION | MB_SYSTEMMODAL)


def command_error(message) -> None:
    _show_message(message, MB_ICONEXCLAMATION | MB_SYSTEMMODAL)


def do_clean() -> None:
    if _wait_for_input:
        print("\nPress Enter to close")
        noerr(sys.stdout.flush)
        noerr(sys.stderr.flush)
        sys.stdin.readline()
        return

    # undo the redirect to file:
    if envbool("XPRA_LOG_TO_FILE", True):
        log_filename = os.environ.get("XPRA_LOG_FILENAME")
        if log_filename and os.path.exists(log_filename):
            noerr(sys.stdout.close)
            sys.stdout = sys.stderr = None



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

cdef extern from "library_types.h":
    cdef enum libraryPropertyType_t:
        MAJOR_VERSION
        MINOR_VERSION
        PATCH_LEVEL

from xpra.codecs.nvidia.nvjpeg.nvjpeg cimport (
    NVJPEG_VER_MAJOR, NVJPEG_VER_MINOR, NVJPEG_VER_PATCH, NVJPEG_VER_BUILD,
    #ERR_STR:
    NVJPEG_STATUS_SUCCESS,
    NVJPEG_STATUS_NOT_INITIALIZED,
    NVJPEG_STATUS_INVALID_PARAMETER,
    NVJPEG_STATUS_BAD_JPEG,
    NVJPEG_STATUS_JPEG_NOT_SUPPORTED,
    NVJPEG_STATUS_ALLOCATOR_FAILURE,
    NVJPEG_STATUS_EXECUTION_FAILED,
    NVJPEG_STATUS_ARCH_MISMATCH,
    NVJPEG_STATUS_INTERNAL_ERROR,
    NVJPEG_STATUS_IMPLEMENTATION_NOT_SUPPORTED,
    #CSS_STR:
    NVJPEG_CSS_444,
    NVJPEG_CSS_422,
    NVJPEG_CSS_420,
    NVJPEG_CSS_440,
    NVJPEG_CSS_411,
    NVJPEG_CSS_410,
    NVJPEG_CSS_GRAY,
    NVJPEG_CSS_UNKNOWN,
    #ENCODING_STR:
    NVJPEG_ENCODING_UNKNOWN,
    NVJPEG_ENCODING_BASELINE_DCT,
    NVJPEG_ENCODING_EXTENDED_SEQUENTIAL_DCT_HUFFMAN,
    NVJPEG_ENCODING_PROGRESSIVE_DCT_HUFFMAN,
    #INPUT_STR:
    NVJPEG_INPUT_RGB,
    NVJPEG_INPUT_BGR,
    NVJPEG_INPUT_RGBI,
    NVJPEG_INPUT_BGRI,
    #OUTPUT_STR:
    NVJPEG_OUTPUT_UNCHANGED,
    NVJPEG_OUTPUT_YUV,
    NVJPEG_OUTPUT_Y,
    NVJPEG_OUTPUT_RGB,
    NVJPEG_OUTPUT_BGR,
    NVJPEG_OUTPUT_RGBI,
    NVJPEG_OUTPUT_BGRI,
    NVJPEG_OUTPUT_FORMAT_MAX,
    #get_version:
    nvjpegGetProperty,
    )


class NVJPEG_Exception(Exception):
    pass

def errcheck(int r, fnname="", *args):
    if r:
        fstr = fnname % (args)
        raise NVJPEG_Exception("%s failed: %s" % (fstr, ERR_STR.get(r, r)))


def get_cuda_version():
    cdef int major_version, minor_version, patch_level
    r = nvjpegGetProperty(MAJOR_VERSION, &major_version)
    errcheck(r, "nvjpegGetProperty MAJOR_VERSION")
    r = nvjpegGetProperty(MINOR_VERSION, &minor_version)
    errcheck(r, "nvjpegGetProperty MINOR_VERSION")
    r = nvjpegGetProperty(PATCH_LEVEL, &patch_level)
    errcheck(r, "nvjpegGetProperty PATCH_LEVEL")
    return (major_version, minor_version, patch_level)

def get_version():
    return (NVJPEG_VER_MAJOR, NVJPEG_VER_MINOR, NVJPEG_VER_PATCH, NVJPEG_VER_BUILD)


ERR_STR = {
    NVJPEG_STATUS_SUCCESS                       : "SUCCESS",
    NVJPEG_STATUS_NOT_INITIALIZED               : "NOT_INITIALIZED",
    NVJPEG_STATUS_INVALID_PARAMETER             : "INVALID_PARAMETER",
    NVJPEG_STATUS_BAD_JPEG                      : "BAD_JPEG",
    NVJPEG_STATUS_JPEG_NOT_SUPPORTED            : "JPEG_NOT_SUPPORTED",
    NVJPEG_STATUS_ALLOCATOR_FAILURE             : "ALLOCATOR_FAILURE",
    NVJPEG_STATUS_EXECUTION_FAILED              : "EXECUTION_FAILED",
    NVJPEG_STATUS_ARCH_MISMATCH                 : "ARCH_MISMATCH",
    NVJPEG_STATUS_INTERNAL_ERROR                : "INTERNAL_ERROR",
    NVJPEG_STATUS_IMPLEMENTATION_NOT_SUPPORTED  : "IMPLEMENTATION_NOT_SUPPORTED",
    }

CSS_STR = {
    NVJPEG_CSS_444  : "444",
    NVJPEG_CSS_422  : "422",
    NVJPEG_CSS_420  : "420",
    NVJPEG_CSS_440  : "440",
    NVJPEG_CSS_411  : "411",
    NVJPEG_CSS_410  : "410",
    NVJPEG_CSS_GRAY : "gray",
    NVJPEG_CSS_UNKNOWN  : "unknown",
    }

ENCODING_STR = {
    NVJPEG_ENCODING_UNKNOWN                         : "unknown",
    NVJPEG_ENCODING_BASELINE_DCT                    : "baseline-dct",
    NVJPEG_ENCODING_EXTENDED_SEQUENTIAL_DCT_HUFFMAN : "extended-sequential-dct-huffman",
    NVJPEG_ENCODING_PROGRESSIVE_DCT_HUFFMAN         : "progressive-dct-huffman",
    }

NVJPEG_INPUT_STR = {
    NVJPEG_INPUT_RGB    : "RGB",
    NVJPEG_INPUT_BGR    : "BGR",
    NVJPEG_INPUT_RGBI   : "RGBI",
    NVJPEG_INPUT_BGRI   : "BGRI",
    }

NVJPEG_OUTPUT_STR = {
    NVJPEG_OUTPUT_UNCHANGED : "UNCHANGED",
    NVJPEG_OUTPUT_YUV       : "YUV",
    NVJPEG_OUTPUT_Y         : "Y",
    NVJPEG_OUTPUT_RGB       : "RGB",
    NVJPEG_OUTPUT_BGR       : "BGR",
    NVJPEG_OUTPUT_RGBI      : "RGBI",
    NVJPEG_OUTPUT_BGRI      : "BGRI",
    #NVJPEG_OUTPUT_FORMAT_MAX
    }



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.gtk.configure.common import run_gui
from xpra.util.config import update_config_env, get_config_env
from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.gtk.widget import label, setfont
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("gstreamer", "util")


SHADOW_BACKENDS: dict[str, Sequence[str]] = {
    "auto": (
        "Automatic runtime detection",
        "this is the default behaviour,",
        "this option should always find a suitable capture strategy",
        "and it may choose not to use a video stream",
    ),
    "x11": (
        "X11 screen capture",
        "copies the X11 server's pixel data,",
        "this option only requires the X11 bindings",
        "incompatible with Wayland sessions, the displays with XWayland will look blank",
    ),
    "xshm": (
        "X11 screen capture via shared memory",
        "identical to `x11` but faster thanks to the XShm extension",
    ),
    "gtk": (
        "GTK screen capture",
        "performance may vary,",
        "this option is not compatible with Wayland displays",
    ),
    "nvfbc": (
        "NVIDIA® Frame Buffer Capture",
        "this requires the proprietary module and libraries",
        "if available, this provides the fastest capture possible",
        "and also supports hardware video compression",
        "this option is only available for shadowing existing X11 sessions",
    ),
    "gstreamer": (
        "GStreamer screen capture",
        "GStreamer will capture the session's contents using an operating system specific source element",
        "the pixel data will be compressed using a stream encoder",
        "eg: h264, hevc, av1, etc",
    ),
    "gdi": (
        "GDI screen capture",
        "Legacy screen capture for MS Windows,",
        "the xpra server can use mixed encodings with this capture option",
    ),
    "pipewire": (
        "GStreamer Pipewire capture",
        "GStreamer pipewire source from the RemoteDesktop interface",
        "the pixel data will be compressed using a stream encoder",
        "eg: h264, hevc, av1, etc",
        "your desktop sessions must support the 'RemoteDesktop' dbus interface",
    ),
}


def _set_labels_text(widgets, *messages):
    for i, widget in enumerate(widgets):
        if i < len(messages):
            widget.set_text(messages[i])
        else:
            widget.set_text("")


class ConfigureGUI(BaseGUIWindow):

    # so we can call check_xshm()
    from xpra.util.system import is_Wayland
    if not is_Wayland():
        from xpra.x11.gtk.display_source import init_gdk_display_source
        init_gdk_display_source()

    def __init__(self, parent: Gtk.Window | None = None):
        self.buttons: list[Gtk.CheckButton] = []
        size = (800, 554)
        super().__init__(
            "Configure Xpra's Shadow Server",
            "shadow.png",
            wm_class=("xpra-configure-shadow-gui", "Xpra Configure Shadow GUI"),
            default_size=size,
            header_bar=(False, False),
            parent=parent,
        )
        self.set_resizable(False)

    def populate(self) -> None:
        self.clear_vbox()
        self.set_box_margin()
        self.add_widget(label("Configure Xpra's Shadow Server", font="sans 20"))
        current_setting = get_config_env("XPRA_SHADOW_BACKEND")
        from xpra.platform.shadow_server import SHADOW_OPTIONS
        for backend, check in SHADOW_OPTIONS.items():
            available = True
            tooltip = ""
            try:
                if not check():
                    available = False
                    tooltip = "not available"
            except ImportError:
                available = False
                tooltip = "not installed or not available"
            details = SHADOW_BACKENDS.get(backend, ())
            if not details:
                description = backend
                tooltip = "unknown backend"
            else:
                description = details[0]
            btn = Gtk.CheckButton(label=description)
            btn.set_tooltip_text(tooltip)
            btn.set_sensitive(available)
            btn.set_active(available and backend == current_setting)
            btn.shadow_backend = backend
            setfont(btn, font="sans 14")
            self.vbox.add(btn)
            for detail in details[1:]:
                lbl = label(detail)
                lbl.set_halign(Gtk.Align.START)
                lbl.set_margin_start(32)
                lbl.set_sensitive(available)
                self.vbox.add(lbl)
            self.buttons.append(btn)
        btn_box = Gtk.HBox(homogeneous=True, spacing=40)
        btn_box.set_vexpand(True)
        btn_box.set_valign(Gtk.Align.END)
        self.vbox.add(btn_box)
        cancel_btn = Gtk.Button.new_with_label("Cancel")
        cancel_btn.connect("clicked", self.dismiss)
        btn_box.add(cancel_btn)
        confirm_btn = Gtk.Button.new_with_label("Confirm")
        confirm_btn.connect("clicked", self.save_shadow)
        confirm_btn.set_sensitive(False)
        btn_box.add(confirm_btn)

        # only enable the confirm button once an option has been chosen,
        # and ensure that there is always one option selected
        def option_toggled(toggled_btn=None, *_args) -> None:
            if toggled_btn and toggled_btn.get_active():
                for button in self.buttons:
                    if button != toggled_btn:
                        button.set_active(False)
            else:
                if not any(button.get_active() for button in self.buttons):
                    self.buttons[0].set_active(True)
            confirm_btn.set_sensitive(any(button.get_active() for button in self.buttons))

        for btn in self.buttons:
            btn.connect("toggled", option_toggled)
        option_toggled()
        self.vbox.show_all()

    def save_shadow(self, *_args) -> None:
        active = [button for button in self.buttons if button.get_active()]
        assert len(active) == 1
        setting = active[0].shadow_backend.lower()
        log.info(f"saving XPRA_SHADOW_BACKEND={setting}")
        update_config_env("XPRA_SHADOW_BACKEND", setting)
        self.dismiss()


def main(_args) -> int:
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from collections import deque
from typing import Any, Deque

from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.net.common import PacketType


class ShellServer(StubServerMixin):
    """
    Mixin for adding `shell` support
    """
    PREFIX = "shell"

    def __init__(self):
        self.counter = 0
        self.commands: Deque[str] = deque(maxlen=10)

    def get_info(self, _source=None) -> dict[str, Any]:
        return {
            ShellServer.PREFIX: {
                "counter": self.counter,
                "last-commands": list(self.commands),
            },
        }

    def get_server_features(self, _source) -> dict[str, Any]:
        return {
            "shell": True,
        }

    def _process_shell_exec(self, proto, packet: PacketType) -> None:
        code = str(packet[1])
        ss = self.get_server_source(proto)
        if ss:
            self.counter += 1
            self.commands.append(code)
            ss.shell_exec(code)

    def init_packet_handlers(self) -> None:
        self.add_packets("shell-exec")



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.common import noop
from xpra.os_util import gi_import, OSX, WIN32
from xpra.util.env import IgnoreWarningsContext, ignorewarnings, envint
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
Pango = gi_import("Pango")
GdkPixbuf = gi_import("GdkPixbuf")

log = Logger("gtk", "util")

FILE_CHOOSER_NATIVE = envint("XPRA_FILE_CHOOSER_NATIVE", int(OSX or WIN32))


def scaled_image(pixbuf, icon_size: int = 0) -> Gtk.Image | None:
    if not pixbuf:
        return None
    if icon_size:
        pixbuf = pixbuf.scale_simple(icon_size, icon_size, GdkPixbuf.InterpType.BILINEAR)
    return Gtk.Image.new_from_pixbuf(pixbuf)


def imagebutton(title, icon=None, tooltip="", clicked_callback: Callable | None = None, icon_size=32,
                default=False, min_size=None, label_color=None, label_font="") -> Gtk.Button:
    button = Gtk.Button(label=title)
    settings = button.get_settings()
    settings.set_property('gtk-button-images', True)
    if icon:
        if icon_size:
            icon = scaled_image(icon, icon_size)
        ignorewarnings(button.set_image, icon)
    if tooltip:
        button.set_tooltip_text(tooltip)
    if min_size:
        button.set_size_request(min_size, min_size)
    if clicked_callback:
        button.connect("clicked", clicked_callback)
    if default:
        button.set_can_default(True)
    if label_color or label_font:
        widget = button
        try:
            alignment = button.get_children()[0]
            b_hbox = alignment.get_children()[0]
            widget = b_hbox.get_children()[1]
        except (IndexError, AttributeError):
            pass
        if label_color:
            modify_fg(widget, label_color)
        if label_font:
            setfont(widget, label_font)
    return button


def modify_fg(widget, color, state=Gtk.StateType.NORMAL) -> None:
    if hasattr(widget, "modify_fg"):
        with IgnoreWarningsContext():
            widget.modify_fg(state, color)


def menuitem(title, image=None, tooltip=None, cb=None) -> Gtk.ImageMenuItem:
    """ Utility method for easily creating an ImageMenuItem """
    menu_item = Gtk.ImageMenuItem()
    menu_item.set_label(title)
    if image:
        ignorewarnings(menu_item.set_image, image)
        # override gtk defaults: we *want* icons:
        settings = menu_item.get_settings()
        settings.set_property('gtk-menu-images', True)
        if hasattr(menu_item, "set_always_show_image"):
            ignorewarnings(menu_item.set_always_show_image, True)
    if tooltip:
        menu_item.set_tooltip_text(tooltip)
    if cb:
        menu_item.connect('activate', cb)
    menu_item.show()
    return menu_item


def label(text="", tooltip="", font="") -> Gtk.Label:
    widget = Gtk.Label(label=text)
    if font:
        setfont(widget, font)
    if tooltip:
        widget.set_tooltip_text(tooltip)
    return widget


def setfont(widget, font=""):
    if font:
        with IgnoreWarningsContext():
            fontdesc = Pango.FontDescription(font)
            widget.modify_font(fontdesc)


def choose_files(parent_window, title, action=Gtk.FileChooserAction.OPEN, action_button=Gtk.STOCK_OPEN,
                 callback=None, file_filter=None, multiple=True) -> list[str]:
    log("choose_files%s", (parent_window, title, action, action_button, callback, file_filter))
    chooser = Gtk.FileChooserDialog(title=title, parent=parent_window, action=action)
    chooser.add_buttons(
        Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
        action_button, Gtk.ResponseType.OK,
    )
    chooser.set_select_multiple(multiple)
    chooser.set_default_response(Gtk.ResponseType.OK)
    if file_filter:
        chooser.add_filter(file_filter)
    response = chooser.run()
    filenames = chooser.get_filenames()
    chooser.hide()
    chooser.destroy()
    if response != Gtk.ResponseType.OK:
        return []
    return filenames


def choose_file(parent_window, title, action=Gtk.FileChooserAction.OPEN, action_button=Gtk.STOCK_OPEN,
                callback=noop, file_filter=None) -> None:
    filenames = choose_files(parent_window, title, action, action_button, callback, file_filter, False)
    if not filenames or len(filenames) != 1:
        return
    filename = filenames[0]
    callback(filename)


orig_pack_start = Gtk.Box.pack_start


def pack_start(self, child, expand=True, fill=True, padding=0):
    orig_pack_start(self, child, expand, fill, padding)


Gtk.Box.pack_start = pack_start


def slabel(text="", tooltip="", font="") -> Gtk.Label:
    lw = label(text, tooltip, font)
    lw.set_margin_start(5)
    lw.set_margin_end(5)
    lw.set_margin_top(2)
    lw.set_margin_bottom(2)
    lw.set_selectable(True)
    lw.set_line_wrap(True)
    return lw


def title_box(label_str: str, tooltip="") -> Gtk.EventBox:
    eb = Gtk.EventBox()
    lbl = slabel(label_str, tooltip=tooltip)
    modify_fg(lbl, Gdk.Color(red=48 * 256, green=0, blue=0))
    al = Gtk.Alignment(xalign=0.0, yalign=0.5, xscale=0.0, yscale=0.0)
    al.set_margin_start(10)
    al.set_margin_end(10)
    al.add(lbl)
    eb.add(al)
    with IgnoreWarningsContext():
        eb.modify_bg(Gtk.StateType.NORMAL, Gdk.Color(red=219 * 256, green=226 * 256, blue=242 * 256))
    return eb


def color_parse(*args) -> Gdk.Color | None:
    v = Gdk.RGBA()
    ok = v.parse(*args)
    if ok:
        return v.to_color()  # pylint: disable=no-member
    with IgnoreWarningsContext():
        ok, v = Gdk.Color.parse(*args)
    if ok:
        return v
    return None


black = color_parse("black")
red = color_parse("red")
white = color_parse("white")


def set_widget_bg_color(widget, is_error=False) -> None:
    with IgnoreWarningsContext():
        widget.modify_base(Gtk.StateType.NORMAL, red if is_error else white)


def set_widget_fg_color(widget, is_error=False) -> None:
    modify_fg(widget, red if is_error else black)



# This file is part of Xpra.
# Copyright (C) 2013 eryksun
# https://stackoverflow.com/questions/15748386/how-to-catch-printer-event-in-python
# License:
# https://creativecommons.org/licenses/by-sa/3.0/legalcode

import ctypes
from ctypes import wintypes
from collections.abc import Sequence

kernel32 = ctypes.WinDLL('kernel32.dll', use_last_error=True)
winspool = ctypes.WinDLL('winspool.drv', use_last_error=True)

# define LPHANDLE, PDWORD, and PWORD for Python 2
if not hasattr(wintypes, 'LPHANDLE'):
    setattr(wintypes, 'LPHANDLE', ctypes.POINTER(wintypes.HANDLE))
if not hasattr(wintypes, 'PDWORD'):
    setattr(wintypes, 'PDWORD', ctypes.POINTER(wintypes.DWORD))
if not hasattr(wintypes, 'PWORD'):
    setattr(wintypes, 'PWORD', ctypes.POINTER(wintypes.WORD))

INFINITE = -1
WAIT_OBJECT_0 = 0x00000000
WAIT_TIMEOUT = 0x00000102
WAIT_FAILED = 0xFFFFFFFF
INVALID_HANDLE_VALUE = wintypes.HANDLE(-1).value

PRINTER_CHANGE_ADD_PRINTER = 0x00000001
PRINTER_CHANGE_SET_PRINTER = 0x00000002
PRINTER_CHANGE_DELETE_PRINTER = 0x00000004
PRINTER_CHANGE_FAILED_CONNECTION_PRINTER = 0x00000008
PRINTER_CHANGE_PRINTER = 0x000000FF

PRINTER_CHANGE_ADD_JOB = 0x00000100
PRINTER_CHANGE_SET_JOB = 0x00000200
PRINTER_CHANGE_DELETE_JOB = 0x00000400
PRINTER_CHANGE_WRITE_JOB = 0x00000800
PRINTER_CHANGE_JOB = 0x0000FF00

PRINTER_CHANGE_ADD_FORM = 0x00010000
PRINTER_CHANGE_SET_FORM = 0x00020000
PRINTER_CHANGE_DELETE_FORM = 0x00040000
PRINTER_CHANGE_FORM = 0x00070000

PRINTER_CHANGE_ADD_PORT = 0x00100000
PRINTER_CHANGE_CONFIGURE_PORT = 0x00200000
PRINTER_CHANGE_DELETE_PORT = 0x00400000
PRINTER_CHANGE_PORT = 0x00700000

PRINTER_CHANGE_ADD_PRINT_PROCESSOR = 0x01000000
PRINTER_CHANGE_DELETE_PRINT_PROCESSOR = 0x04000000
PRINTER_CHANGE_PRINT_PROCESSOR = 0x07000000
PRINTER_CHANGE_SERVER = 0x08000000  # NT 6.1+

PRINTER_CHANGE_ADD_PRINTER_DRIVER = 0x10000000
PRINTER_CHANGE_SET_PRINTER_DRIVER = 0x20000000
PRINTER_CHANGE_DELETE_PRINTER_DRIVER = 0x40000000
PRINTER_CHANGE_PRINTER_DRIVER = 0x70000000

PRINTER_CHANGE_ALL = 0x7F77FFFF
PRINTER_CHANGE_TIMEOUT = 0x80000000

PRINTER_NOTIFY_CATEGORY_ALL = 0x00
PRINTER_NOTIFY_CATEGORY_3D = 0x01

PRINTER_NOTIFY_TYPE = 0x00
JOB_NOTIFY_TYPE = 0x01

PRINTER_NOTIFY_FIELD_SERVER_NAME = 0x00  # not supported
PRINTER_NOTIFY_FIELD_PRINTER_NAME = 0x01
PRINTER_NOTIFY_FIELD_SHARE_NAME = 0x02
PRINTER_NOTIFY_FIELD_PORT_NAME = 0x03
PRINTER_NOTIFY_FIELD_DRIVER_NAME = 0x04
PRINTER_NOTIFY_FIELD_COMMENT = 0x05
PRINTER_NOTIFY_FIELD_LOCATION = 0x06
PRINTER_NOTIFY_FIELD_DEVMODE = 0x07
PRINTER_NOTIFY_FIELD_SEPFILE = 0x08
PRINTER_NOTIFY_FIELD_PRINT_PROCESSOR = 0x09
PRINTER_NOTIFY_FIELD_PARAMETERS = 0x0A
PRINTER_NOTIFY_FIELD_DATATYPE = 0x0B
PRINTER_NOTIFY_FIELD_SECURITY_DESCRIPTOR = 0x0C
PRINTER_NOTIFY_FIELD_ATTRIBUTES = 0x0D
PRINTER_NOTIFY_FIELD_PRIORITY = 0x0E
PRINTER_NOTIFY_FIELD_DEFAULT_PRIORITY = 0x0F
PRINTER_NOTIFY_FIELD_START_TIME = 0x10
PRINTER_NOTIFY_FIELD_UNTIL_TIME = 0x11
PRINTER_NOTIFY_FIELD_STATUS = 0x12
PRINTER_NOTIFY_FIELD_STATUS_STRING = 0x13  # not supported
PRINTER_NOTIFY_FIELD_CJOBS = 0x14
PRINTER_NOTIFY_FIELD_AVERAGE_PPM = 0x15
PRINTER_NOTIFY_FIELD_TOTAL_PAGES = 0x16  # not supported
PRINTER_NOTIFY_FIELD_PAGES_PRINTED = 0x17  # not supported
PRINTER_NOTIFY_FIELD_TOTAL_BYTES = 0x18  # not supported
PRINTER_NOTIFY_FIELD_BYTES_PRINTED = 0x19  # not supported
PRINTER_NOTIFY_FIELD_OBJECT_GUID = 0x1A
PRINTER_NOTIFY_FIELD_FRIENDLY_NAME = 0x1B  # NT 6.0+
PRINTER_NOTIFY_FIELD_BRANCH_OFFICE_PRINTING = 0x1C  # NT 6.2+

JOB_NOTIFY_FIELD_PRINTER_NAME = 0x00
JOB_NOTIFY_FIELD_MACHINE_NAME = 0x01
JOB_NOTIFY_FIELD_PORT_NAME = 0x02
JOB_NOTIFY_FIELD_USER_NAME = 0x03
JOB_NOTIFY_FIELD_NOTIFY_NAME = 0x04
JOB_NOTIFY_FIELD_DATATYPE = 0x05
JOB_NOTIFY_FIELD_PRINT_PROCESSOR = 0x06
JOB_NOTIFY_FIELD_PARAMETERS = 0x07
JOB_NOTIFY_FIELD_DRIVER_NAME = 0x08
JOB_NOTIFY_FIELD_DEVMODE = 0x09
JOB_NOTIFY_FIELD_STATUS = 0x0A
JOB_NOTIFY_FIELD_STATUS_STRING = 0x0B
JOB_NOTIFY_FIELD_SECURITY_DESCRIPTOR = 0x0C  # not supported
JOB_NOTIFY_FIELD_DOCUMENT = 0x0D
JOB_NOTIFY_FIELD_PRIORITY = 0x0E
JOB_NOTIFY_FIELD_POSITION = 0x0F
JOB_NOTIFY_FIELD_SUBMITTED = 0x10
JOB_NOTIFY_FIELD_START_TIME = 0x11
JOB_NOTIFY_FIELD_UNTIL_TIME = 0x12
JOB_NOTIFY_FIELD_TIME = 0x13
JOB_NOTIFY_FIELD_TOTAL_PAGES = 0x14
JOB_NOTIFY_FIELD_PAGES_PRINTED = 0x15
JOB_NOTIFY_FIELD_TOTAL_BYTES = 0x16
JOB_NOTIFY_FIELD_BYTES_PRINTED = 0x17
JOB_NOTIFY_FIELD_REMOTE_JOB_ID = 0x18

PRINTER_NOTIFY_OPTIONS_REFRESH = 0x01
PRINTER_NOTIFY_INFO_DISCARDED = 0x01

JOB_STATUS_PAUSED = 0x00000001
JOB_STATUS_ERROR = 0x00000002
JOB_STATUS_DELETING = 0x00000004
JOB_STATUS_SPOOLING = 0x00000008
JOB_STATUS_PRINTING = 0x00000010
JOB_STATUS_OFFLINE = 0x00000020
JOB_STATUS_PAPEROUT = 0x00000040
JOB_STATUS_PRINTED = 0x00000080
JOB_STATUS_DELETED = 0x00000100
JOB_STATUS_BLOCKED_DEVQ = 0x00000200
JOB_STATUS_USER_INTERVENTION = 0x00000400
JOB_STATUS_RESTART = 0x00000800
JOB_STATUS_COMPLETE = 0x00001000
JOB_STATUS_RETAINED = 0x00002000
JOB_STATUS_RENDERING_LOCALLY = 0x00004000
JOB_STATUS_ALL = 0x00007FFF

JOB_STATUS_STRING = {
    JOB_STATUS_PAUSED: 'PAUSED',
    JOB_STATUS_ERROR: 'ERROR',
    JOB_STATUS_DELETING: 'DELETING',
    JOB_STATUS_SPOOLING: 'SPOOLING',
    JOB_STATUS_PRINTING: 'PRINTING',
    JOB_STATUS_OFFLINE: 'OFFLINE',
    JOB_STATUS_PAPEROUT: 'PAPEROUT',
    JOB_STATUS_PRINTED: 'PRINTED',
    JOB_STATUS_DELETED: 'DELETED',
    JOB_STATUS_BLOCKED_DEVQ: 'BLOCKED_DEVQ',
    JOB_STATUS_USER_INTERVENTION: 'USER_INTERVENTION',
    JOB_STATUS_RESTART: 'RESTART',
    JOB_STATUS_COMPLETE: 'COMPLETE',
    JOB_STATUS_RETAINED: 'RETAINED',
    JOB_STATUS_RENDERING_LOCALLY: 'RENDERING_LOCALLY'}


class SYSTEMTIME(ctypes.Structure):
    _fields_ = (('wYear', wintypes.WORD),
                ('wMonth', wintypes.WORD),
                ('wDayOfWeek', wintypes.WORD),
                ('wDay', wintypes.WORD),
                ('wHour', wintypes.WORD),
                ('wMinute', wintypes.WORD),
                ('wSecond', wintypes.WORD),
                ('wMilliseconds', wintypes.WORD))

    @property
    def as_datetime(self):
        from datetime import datetime
        return datetime(self.wYear, self.wMonth, self.wDay,
                        self.wHour, self.wMinute, self.wSecond,
                        self.wMilliseconds * 1000)


# noinspection PyTypeChecker
class PRINTER_NOTIFY_OPTIONS_TYPE(ctypes.Structure):
    _fields_ = (('Type', wintypes.WORD),
                ('Reserved0', wintypes.WORD),
                ('Reserved1', wintypes.DWORD),
                ('Reserved2', wintypes.DWORD),
                ('Count', wintypes.DWORD),
                ('_pFields', wintypes.PWORD))

    def __init__(self, Type=JOB_NOTIFY_TYPE, pFields=None):
        super().__init__(Type)
        if pFields is not None:
            self.pFields = pFields

    @property
    def pFields(self):
        ptr_t = ctypes.POINTER(wintypes.WORD * self.Count)
        return ptr_t(self._pFields.contents)[0]

    @pFields.setter
    def pFields(self, pFields):
        self.Count = len(pFields)
        self._pFields = pFields


PPRINTER_NOTIFY_OPTIONS_TYPE = ctypes.POINTER(PRINTER_NOTIFY_OPTIONS_TYPE)


class PRINTER_NOTIFY_OPTIONS(ctypes.Structure):
    _fields_ = (('Version', wintypes.DWORD),
                ('Flags', wintypes.DWORD),
                ('Count', wintypes.DWORD),
                ('_pTypes', PPRINTER_NOTIFY_OPTIONS_TYPE))

    def __init__(self, Flags=0, pTypes=None):
        super().__init__(2, Flags)
        if pTypes is not None:
            self.pTypes = pTypes

    @property
    def pTypes(self):
        ptr_t = ctypes.POINTER(PRINTER_NOTIFY_OPTIONS_TYPE * self.Count)
        return ptr_t(self._pTypes.contents)[0]

    @pTypes.setter
    def pTypes(self, types):
        if isinstance(types, PRINTER_NOTIFY_OPTIONS_TYPE):
            self.Count = 1
            self._pTypes = ctypes.pointer(types)
        else:
            self.Count = len(types)
            self._pTypes = types


PPRINTER_NOTIFY_OPTIONS = ctypes.POINTER(PRINTER_NOTIFY_OPTIONS)


class PRINTER_NOTIFY_INFO_DATA(ctypes.Structure):
    class _NOTIFY_DATA(ctypes.Union):
        class _DATA(ctypes.Structure):
            _fields_ = (('cbBuf', wintypes.DWORD),
                        ('pBuf', wintypes.LPVOID))

        # noinspection PyTypeChecker
        _fields_ = (('adwData', wintypes.DWORD * 2),
                    ('Data', _DATA))

    _fields_ = (('Type', wintypes.WORD),
                ('Field', wintypes.WORD),
                ('Reserved', wintypes.DWORD),
                ('Id', wintypes.DWORD),
                ('_NotifyData', _NOTIFY_DATA))

    @property
    def _data_as_string(self):
        if self._NotifyData.Data.pBuf:
            return ctypes.c_wchar_p(self._NotifyData.Data.pBuf).value
        return ""

    @property
    def _data_as_datetime(self):
        if self._NotifyData.Data.pBuf:
            t = SYSTEMTIME.from_address(self._NotifyData.Data.pBuf)
        else:
            t = SYSTEMTIME()
        return t.as_datetime

    @property
    def NotifyData(self):
        if self.Type == JOB_NOTIFY_TYPE:
            if self.Field == JOB_NOTIFY_FIELD_PRINTER_NAME:
                return 'job_printer_name', self._data_as_string
            if self.Field == JOB_NOTIFY_FIELD_MACHINE_NAME:
                return 'job_machine_name', self._data_as_string
            if self.Field == JOB_NOTIFY_FIELD_USER_NAME:
                return 'job_user_name', self._data_as_string
            elif self.Field == JOB_NOTIFY_FIELD_STATUS:
                return 'job_status', self._NotifyData.adwData[0]
            elif self.Field == JOB_NOTIFY_FIELD_DOCUMENT:
                return 'job_document', self._data_as_string
            elif self.Field == JOB_NOTIFY_FIELD_PRIORITY:
                return 'job_priority', self._NotifyData.adwData[0]
            elif self.Field == JOB_NOTIFY_FIELD_POSITION:
                return 'job_position', self._NotifyData.adwData[0]
            elif self.Field == JOB_NOTIFY_FIELD_SUBMITTED:
                return 'job_submitted', self._data_as_datetime
            elif self.Field == JOB_NOTIFY_FIELD_PAGES_PRINTED:
                return 'job_pages_printed', self._NotifyData.adwData[0]
            elif self.Field == JOB_NOTIFY_FIELD_BYTES_PRINTED:
                return 'job_bytes_printed', self._NotifyData.adwData[0]
        # else return a copy of NotifyData
        data = self._NOTIFY_DATA.from_buffer_copy(self._NotifyData)
        if data.Data.pBuf:
            buf_t = ctypes.c_char * data.Data.cbBuf
            buf_src = buf_t.from_address(data.Data.pBuf)
            buf_cpy = buf_t.from_buffer_copy(buf_src)
            buf_ptr = ctypes.c_void_p(ctypes.addressof(buf_cpy))
            data.Data.pBuf = buf_ptr
        return (self.Type, self.Field), data


class PRINTER_NOTIFY_INFO(ctypes.Structure):
    _fields_ = (('Version', wintypes.DWORD),
                ('Flags', wintypes.DWORD),
                ('Count', wintypes.DWORD),
                ('_aData', PRINTER_NOTIFY_INFO_DATA * 1))

    @property
    def aData(self):
        ptr_t = ctypes.POINTER(PRINTER_NOTIFY_INFO_DATA * self.Count)
        return ptr_t(self._aData[0])[0]


PPRINTER_NOTIFY_INFO = ctypes.POINTER(PRINTER_NOTIFY_INFO)
PPPRINTER_NOTIFY_INFO = ctypes.POINTER(PPRINTER_NOTIFY_INFO)


class PPRINTER_NOTIFY_INFO_GC(PPRINTER_NOTIFY_INFO):
    """PRINTER_NOTIFY_INFO * that Windows deallocates"""
    _type_ = PRINTER_NOTIFY_INFO
    _freed = False

    def __del__(self, FreePrinterNotifyInfo=winspool.FreePrinterNotifyInfo):
        if self and not self._freed:
            FreePrinterNotifyInfo(self)
            self._freed = True


def check_bool(result, _func, args):
    if not result:
        raise ctypes.WinError(ctypes.get_last_error())
    return args


def check_ihv(result, _func, args):
    if result == INVALID_HANDLE_VALUE:
        raise ctypes.WinError(ctypes.get_last_error())
    return args


def check_idv(result, _func, args):
    if result == WAIT_FAILED:
        raise ctypes.WinError(ctypes.get_last_error())
    return args


winspool.OpenPrinterW.errcheck = check_bool
winspool.OpenPrinterW.argtypes = (
    wintypes.LPWSTR,  # _In_  pPrinterName
    wintypes.LPHANDLE,  # _Out_ phPrinter
    wintypes.LPVOID)  # _In_  pDefault

winspool.ClosePrinter.errcheck = check_bool
winspool.ClosePrinter.argtypes = (
    wintypes.HANDLE,)  # _In_ hPrinter

winspool.FindFirstPrinterChangeNotification.errcheck = check_ihv
winspool.FindFirstPrinterChangeNotification.restype = wintypes.HANDLE
winspool.FindFirstPrinterChangeNotification.argtypes = (
    wintypes.HANDLE,  # _In_ hPrinter
    wintypes.DWORD,  # fdwFilter
    wintypes.DWORD,  # fdwOptions
    PPRINTER_NOTIFY_OPTIONS)  # _In_opt_ pPrinterNotifyOptions

winspool.FindNextPrinterChangeNotification.errcheck = check_bool
winspool.FindNextPrinterChangeNotification.argtypes = (
    wintypes.HANDLE,  # _In_      hChange
    wintypes.PDWORD,  # _Out_opt_ pdwChange
    PPRINTER_NOTIFY_OPTIONS,  # _In_opt_  pPrinterNotifyOptions
    PPPRINTER_NOTIFY_INFO)  # _Out_opt_ ppPrinterNotifyInfo

winspool.FindClosePrinterChangeNotification.errcheck = check_bool
winspool.FindClosePrinterChangeNotification.argtypes = (
    wintypes.HANDLE,)  # _In_ hChange

winspool.FreePrinterNotifyInfo.errcheck = check_bool
winspool.FreePrinterNotifyInfo.argtypes = (
    PPRINTER_NOTIFY_INFO,)  # _In_ pPrinterNotifyInfo

kernel32.WaitForSingleObject.errcheck = check_idv
kernel32.WaitForSingleObject.restype = wintypes.DWORD
kernel32.WaitForSingleObject.argtypes = (
    wintypes.HANDLE,  # _In_ hHandle
    wintypes.DWORD)  # _In_ dwMilliseconds


def wait_for_print_job(printer_filter=PRINTER_CHANGE_ADD_JOB,
                       timeout: int = INFINITE,
                       printer_name: str | None = None) -> int:
    if timeout != INFINITE:
        timeout = int(timeout * 1000)
    hPrinter = wintypes.HANDLE()
    dwChange = wintypes.DWORD()
    winspool.OpenPrinterW(printer_name, ctypes.byref(hPrinter), None)
    try:
        hChange = winspool.FindFirstPrinterChangeNotification(hPrinter, printer_filter, 0, None)
        try:
            if kernel32.WaitForSingleObject(hChange, timeout) != WAIT_OBJECT_0:
                return 0
            winspool.FindNextPrinterChangeNotification(hChange, ctypes.byref(dwChange), None, None)
            return int(dwChange.value)
        finally:
            winspool.FindClosePrinterChangeNotification(hChange)
    finally:
        winspool.ClosePrinter(hPrinter)


DEFAULT_FIELDS: Sequence[int] = (
    JOB_NOTIFY_FIELD_PRINTER_NAME,
    JOB_NOTIFY_FIELD_STATUS,
    JOB_NOTIFY_FIELD_DOCUMENT,
    JOB_NOTIFY_FIELD_PRIORITY,
    JOB_NOTIFY_FIELD_POSITION,
    JOB_NOTIFY_FIELD_SUBMITTED,
)


def wait_for_print_job_info(fields=DEFAULT_FIELDS,
                            timeout=INFINITE,
                            printer_name=None) -> list[list]:
    if timeout != INFINITE:
        timeout = int(timeout * 1000)
    hPrinter = wintypes.HANDLE()
    # noinspection PyCallingNonCallable,PyTypeChecker
    fields = (wintypes.WORD * len(fields))(*fields)
    opt = PRINTER_NOTIFY_OPTIONS(
        pTypes=PRINTER_NOTIFY_OPTIONS_TYPE(
            Type=JOB_NOTIFY_TYPE, pFields=fields))
    pinfo = PPRINTER_NOTIFY_INFO_GC()  # note: GC subclass
    result: list[list] = []
    winspool.OpenPrinterW(printer_name, ctypes.byref(hPrinter), None)
    try:
        hChange = winspool.FindFirstPrinterChangeNotification(
            hPrinter, 0, 0, ctypes.byref(opt))
        try:
            if kernel32.WaitForSingleObject(hChange, timeout) != WAIT_OBJECT_0:
                return result
            winspool.FindNextPrinterChangeNotification(hChange, None, None, ctypes.byref(pinfo))
            for data in pinfo[0].aData:
                if data.Type != JOB_NOTIFY_TYPE:
                    continue
                nd = (data.Id,) + data.NotifyData
                result.append(nd)
            return result
        finally:
            winspool.FindClosePrinterChangeNotification(hChange)
    finally:
        winspool.ClosePrinter(hPrinter)


def job_status(status, nfmt='%#010x'):
    if status == 0:
        return []
    strings = []
    for state, string in JOB_STATUS_STRING.items():
        if status & state:
            strings.append(string)
            status &= ~state
            if not status:
                break
    if status:
        strings.append(nfmt % status)
    return strings


def job_status_string(status, nfmt='%#010x'):
    return ','.join(job_status(status, nfmt))
